# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Utilities for manipulating xarray Datasets and DataArrays."""

# Python Standard Libraries
import logging

# Third Party Installed Libraries
import numpy
from pyresample import utils
import xarray

LOG = logging.getLogger(__name__)

# scipy.interpolate.griddata requires at least 4 points. Don't bother
# plotting if fewer than 4.
MIN_POINTS = 4


def get_lat_lon_points(checklat, checklon, diff, sect_xarray, varname, drop=False):
    """Pull values from xarray Datasets in specified geographic location.

    Return points a given distance around a specified lat/lon location,
    from xarray Datasets.

    Parameters
    ----------
    checklat : float
        latitude of interest
    checklon : float
        longitude of interest
    diff : float
        check +- diff of latitude and longitude
    sect_xarray : Dataset
        xarray dataset containing 'latitude' 'longitude' and varname variables
    varname : str
        variable name of data array to use for returning data values

    Returns
    -------
    float, float, int
        * min value in range
        * max value in range
        * and number of points in range
    """
    xinds = (
        (sect_xarray["longitude"] > checklon - diff)
        & (sect_xarray["longitude"] < checklon + diff)
        & (sect_xarray["latitude"] > checklat - diff)
        & (sect_xarray["latitude"] < checklat + diff)
    )
    import numpy

    if drop is True:
        retval = sect_xarray[varname].where(xinds, drop=drop).size
    else:
        retval = numpy.ma.count(sect_xarray[varname].where(xinds).to_masked_array())
    return (
        sect_xarray[varname].where(xinds).min(),
        sect_xarray[varname].where(xinds).max(),
        retval,
    )


def get_lat_lon_points_numpy(
    checklat, checklon, diff, lat_array, lon_array, data_array
):
    """Pull values from numpy arrays in specified geographic location.

    Return points a given distance around a specified lat/lon location,
    from numpy arrays.

    Parameters
    ----------
    checklat : float
        latitude of interest
    checklon : float
        longitude of interest
    diff : float
        check +- diff of latitude and longitude
    lat_array : ndarray
        numpy ndarray of latitude locations - same shape as lon_array and data_array
    lon_array : ndarray
        numpy ndarray of longitude locations - same shape as lat_array and data_array
    data_array : ndarray
        numpy ndarray data values - same shape as lat_array and lon_array

    Returns
    -------
    float, float, int
        * min value in range
        * max value in range
        * and number of points in range
    """
    import numpy

    inds = numpy.logical_and(
        lat_array > checklat - diff, lat_array < checklat + diff
    ) & numpy.logical_and(lon_array > checklon - diff, lon_array < checklon + diff)

    return (
        data_array[inds].min(),
        data_array[inds].max(),
        numpy.ma.count(data_array[inds]),
    )


def sector_xarray_temporal(
    full_xarray, mindt, maxdt, varnames, verbose=False, drop=False
):
    """Sector an xarray object temporally.  If full_xarray is None, return None.

    Parameters
    ----------
    full_xarray : xarray.Dataset
        xarray object to sector temporally
    mindt : datetime.datetime
        minimum datetime of desired data
    maxdt : datetime.datetime
        maximum datetime of desired data
    varnames : list of str
        list of variable names that should be sectored based on 'time',
        mindt, maxdt

    Returns
    -------
    xarray Dataset, or None
        * if full_xarray is None, return None
        * return full original xarray object if 'time' is not included in
          varnames list
        * else, return sectored xarray object with only the desired times,
          specified by mindt and maxdt
    """
    import numpy

    if full_xarray is None:
        LOG.info(
            "    full_xarray is None - not attempting to sector temporally, "
            "returning None"
        )
        return None

    import xarray

    time_xarray = xarray.Dataset()
    time_xarray.attrs = full_xarray.attrs.copy()

    if "time" not in varnames:
        LOG.info(
            "    time variable not included in list - not temporally sectoring, "
            "returning all data"
        )
        return full_xarray

    for varname in varnames:
        # To masked array can have a performance hit for dask arrays,
        # do not use them unnecessarily.
        # good_speeds = numpy.ma.count(full_xarray[varname].to_masked_array())
        if verbose:
            LOG.info(
                "    STARTED TIME WITH %s points for %s",
                full_xarray[varname].size,
                varname,
            )
    mindt64 = numpy.datetime64(mindt)
    maxdt64 = numpy.datetime64(maxdt)

    xarray_time_mask = (full_xarray["time"] > mindt64) & (full_xarray["time"] < maxdt64)

    time_inds = numpy.where(xarray_time_mask)

    if len(time_inds[0]) == 0:
        LOG.warning(
            "    NO TIME DATA between %s and %s for any vars, skipping", mindt, maxdt
        )
        return None

    min_y = time_inds[0].min()
    max_y = time_inds[0].max() + 1
    if len(time_inds) == 2:
        min_x = time_inds[1].min()
        max_x = time_inds[1].max() + 1

    # covg = False
    # final_good_points = 0
    for varname in varnames:
        if drop is True:
            if len(time_inds) == 2:
                # time_xarray[varname] =
                #                full_xarray[varname].where(xarray_time_mask, drop=drop)
                time_xarray[varname] = full_xarray[varname][min_y:max_y, min_x:max_x]
            else:
                # time_xarray[varname] =
                #                full_xarray[varname].where(xarray_time_mask, drop=drop)
                time_xarray[varname] = full_xarray[varname][min_y:max_y]
        else:
            time_xarray[varname] = full_xarray[varname].where(xarray_time_mask)

        # To masked array can have a performance hit for dask arrays,
        # do not use them unnecessarily.
        # good_speeds = numpy.ma.count(time_xarray[varname].to_masked_array())
        # if good_speeds > final_good_points:
        #     final_good_points = good_speeds
        # if good_speeds < MIN_POINTS:
        #     if verbose:
        #         LOG.warning('    INSUFFICIENT TIME DATA between %s and %s for var %s,
        #                     %s points', mindt, maxdt, varname, good_speeds)
        # else:
        #     if verbose:
        #         LOG.info('    MATCHED TIME %s points for var %s', good_speeds,
        #                  varname)
        #     covg = True

    # if not covg:
    if time_xarray["time"].size == 0:
        LOG.warning(
            "    INSUFFICIENT TIME DATA between %s and %s for any vars, skipping",
            mindt,
            maxdt,
        )
        return None
    LOG.info(
        "    SUFFICIENT TIME DATA between %s and %s for any var %s points",
        mindt,
        maxdt,
        time_xarray["time"].size,
    )
    return time_xarray


def sector_xarray_spatial(
    full_xarray,
    area_def,
    varnames,
    lon_pad=3,
    lat_pad=0,
    verbose=False,
    drop=False,
):
    """Sector an xarray object spatially.  If full_xarray is None, return None.

    Parameters
    ----------
    full_xarray : xarray.Dataset
        xarray object to sector spatially
    area_def: pyresample.AreaDefinition
        The requested region to sector spatially from the full_xarray.
        If None, no spatial sectoring required.
    varnames : list of str
        list of variable names that should be sectored based on 'time'
    drop : bool
        Specify whether to remove points with no coverage (rather than masking)

    Returns
    -------
    xarray.Dataset
        * if full_xarray is None, return None,
        * else return resulting xarray Dataset.
    """
    if full_xarray is None:
        if verbose:
            LOG.info(
                "    full_xarray is None - not attempting to sector spatially, "
                "returning None"
            )
        return None

    sector_xarray = xarray.Dataset()
    sector_xarray.attrs = full_xarray.attrs.copy()

    if verbose:
        LOG.info("    Padding longitudes")
    if verbose:
        LOG.info("    Padding latitudes")

    min_lon, min_lat, max_lon, max_lat = get_minmax_latlon_from_area_def(
        area_def, lon_pad, lat_pad
    )

    if verbose:
        LOG.info("    Wrapping longitudes")
    lons = utils.wrap_longitudes(full_xarray["longitude"])

    if verbose:
        LOG.info("    Handling dateline")
    if lons.max() > 179.5 and lons.min() < -179.5 and max_lon > 0 and min_lon > 0:
        lons = xarray.where(
            full_xarray["longitude"] < 0,
            full_xarray["longitude"] + 360,
            full_xarray["longitude"],
        )
    lats = full_xarray["latitude"]

    # To masked array can have a performance hit for dask arrays,
    # do not use them unnecessarily
    # for varname in varnames:
    #     good_speeds = numpy.ma.count(full_xarray[varname].to_masked_array())
    #     if verbose:
    #         LOG.info('    STARTED SPATIAL WITH %s points for %s', good_speeds,
    #                  varname)

    if verbose:
        LOG.info(
            "    Getting appropriate sector area lon %s to %s lat %s to %s, data "
            "minlon %s, maxlon %s, minlat %s, maxlat %s, %s points",
            min_lon,
            max_lon,
            min_lat,
            max_lat,
            lons.min().data,
            lons.max().data,
            lats.min().data,
            lats.max().data,
            lats.size,
        )
        # lons.min().data, lons.max().data, lats.min().data, lats.max().data,
        # good_speeds)

    xarray_sector_mask = (
        (lons > min_lon) & (lons < max_lon) & (lats > min_lat) & (lats < max_lat)
    )

    sector_inds = numpy.where(xarray_sector_mask)

    if not len(sector_inds[0]):
        LOG.warning(
            "    NO SPATIAL DATA between %0.2f and %0.2f lon and %0.2f and %0.2f lat",
            min_lon,
            max_lon,
            min_lat,
            max_lat,
        )
        return None

    min_y = sector_inds[0].min()
    max_y = sector_inds[0].max() + 1
    if len(sector_inds) == 2:
        min_x = sector_inds[1].min()
        max_x = sector_inds[1].max() + 1

    # covg = False
    # final_good_points = 0
    for varname in varnames:
        if drop is True:
            if len(sector_inds) == 2:
                # sector_xarray[varname] =
                #              full_xarray[varname].where(xarray_sector_mask, drop=drop)
                sector_xarray[varname] = full_xarray[varname][min_y:max_y, min_x:max_x]
            else:
                # sector_xarray[varname] =
                #              full_xarray[varname].where(xarray_sector_mask, drop=drop)
                sector_xarray[varname] = full_xarray[varname][min_y:max_y]
        else:
            sector_xarray[varname] = full_xarray[varname].where(xarray_sector_mask)

        # To masked array can have a performance hit for dask arrays,
        # do not use them unnecessarily
        # good_speeds = numpy.ma.count(sector_xarray[varname].to_masked_array())
        # if good_speeds > final_good_points:
        #     final_good_points = good_speeds

        # if sector_xarray[varname].size < MIN_POINTS or good_speeds < MIN_POINTS:
        #     if verbose:
        #         LOG.warning('    INSUFFICIENT SPATIAL DATA between %0.2f and %0.2f lon
        #                     and %0.2f and %0.2f lat, varname %s, %s points',
        #                     extent_lonlat[0], extent_lonlat[2], extent_lonlat[1],
        #                     extent_lonlat[3], varname, good_speeds)
        # else:
        #     if verbose:
        #         LOG.info('    MATCHED SPATIAL %s points for var %s after location
        #                  sectoring', good_speeds, varname)
        #     covg = True

    # if covg:
    if sector_xarray["latitude"].size == 0:
        LOG.warning(
            "    OVERALL INSUFFICIENT SPATIAL DATA between %0.2f and %0.2f lon"
            " and %0.2f and %0.2f lat",
            min_lat,
            max_lon,
            min_lat,
            max_lat,
        )
        return None

    LOG.info(
        "    OVERALL SUFFICIENT SPATIAL DATA between %0.2f and %0.2f lon and %0.2f"
        " and %0.2f lat %0.2f points",
        min_lon,
        max_lon,
        min_lat,
        max_lat,
        sector_xarray["latitude"].size,
    )
    # extent_lonlat[0], extent_lonlat[2], extent_lonlat[1], extent_lonlat[3],
    # final_good_points)
    return sector_xarray


def get_minmax_latlon_from_area_def(area_def, lon_pad, lat_pad):
    """Retrieve the Min/Max Lat/Lon from the provided area_def.

    Given a PyResample AreaDefinition Object, retrieve the minimum and maximum
    latitude and longitude values that should be used to sector the xarray efficiently.

    If the provided area def encapsulates either of the poles, set the min/max latitude
    value to -90/90 degrees accordingly, and match the other latitude value to what was
    provided in the area_def bounds.

    For longitudes, if the sector doesn't contain a pole, wrap them to ensure they exist
    in the 0-360 range, otherwise set to -180, 180.

    Parameters
    ----------
    area_def: PyResample AreaDefinition()
        - The area_definition that defines the bounds of the data we will be sectoring.
    lon_pad: float
        - The amount (in degrees) of padding to be added to the longitude bounds.
    lat_pad: float
        - The amount (in degrees) of padding to be added to the latitude bounds.

    Returns
    -------
    min_lon, min_lat, max_lon, max_lat: float
        - These degree (float) values will be returned in the order listed above.
    """
    extent_lonlat = list(
        area_def.area_extent_ll
    )  # [min_lon, min_lat, max_lon, max_lat]
    min_lon, max_lon = -180, 180

    # Check if the area_def includes either North or South Pole. If it does, we have to
    # manually adjust the lat_lon extent to reflect the correct extent for both lat/lon.
    # Otherwise go about business as usual.
    LOG.info("Checking if area_def includes the North Pole.")
    if point_in_area_def(area_def, 0, 90):
        LOG.info("North Pole exists within area_def.")
        min_lat = extent_lonlat[1] - lat_pad
        max_lat = 90
        return min_lon, min_lat, max_lon, max_lat

    LOG.info("North Pole not in area_def, checking South Pole.")
    if point_in_area_def(area_def, 0, -90):
        LOG.info("South Pole exists within area_def.")
        min_lat = -90
        max_lat = extent_lonlat[3] + lat_pad
        return min_lon, min_lat, max_lon, max_lat

    LOG.info(
        "Neither North or South Pole are in area def, using custom bounds.",
    )
    # Area Definition doesn't include a pole, get the natural bounds from the area_def.
    # convert extent longitude to be within [-180, 180] range
    min_lon = utils.wrap_longitudes(extent_lonlat[0]) - lon_pad
    max_lon = utils.wrap_longitudes(extent_lonlat[2]) + lon_pad
    if min_lon > max_lon and max_lon < 0:
        max_lon = max_lon + 360
    min_lat = extent_lonlat[1] - lat_pad
    max_lat = extent_lonlat[3] + lat_pad
    # Make sure we don't extend latitudes beyond -90 / +90
    if min_lat < -90.0:
        min_lat = -90.0
    if max_lat > 90.0:
        max_lat = 90.0
    return min_lon, min_lat, max_lon, max_lat


def point_in_area_def(area_def, lon, lat):
    """Determine whether or not the provided point (in degrees) is within area_def.

    Parameters
    ----------
    area_def: PyResample AreaDefinition()
        - The area_definition that defines the bounds of the data we will be sectoring.
    lon: float
        - The longitude value in degrees.
    lat: float
        - The latitude value in degrees.

    Returns
    -------
    bool:
        - Whether or not the point is contained in the area definition.
    """
    try:
        area_def.get_array_indices_from_lonlat(lon, lat)
        return True
    except ValueError:
        return False


def sector_xarray_dataset(
    full_xarray,
    area_def,
    varnames,
    lon_pad=3,
    lat_pad=0,
    verbose=False,
    hours_before_sector_time=18,
    hours_after_sector_time=6,
    drop=False,
    window_start_time=None,
    window_end_time=None,
):
    """Use the xarray to appropriately sector out data by lat/lon and time.

    Parameters
    ----------
    full_xarray: xarray.Dataset
        The full input xarray Dataset
    area_def: pyresample.AreaDefinition
        The requested region to sector spatially from the full_xarray.
        If None, no spatial sectoring required.
    varnames: list of str
        List of variable names within full_xarray to include in the final
        xarray Dataset
    lon_pad: float, default=3
        Amount of padding to include around the longitude dimension
    lat_pad: float, default=0
        Amount of padding to include around the latitude dimension
    verbose: bool, default=False
        If True, include log output, otherwise return silently.
    hours_before_sector_time: float, default=18
        For dynamic sectors, number of hours before sector start time
        to include when sectoring temporally.
    hours_after_sector_time: float, default=6
        For dynamic sectors, number of hours after sector end time
        to include when sectoring temporally.
    drop: bool, default=False
        If True, drop data outside range, temporally and spatially.
        If False, mask outside range.
    window_start_time: datetime.datetime, default=None
        If specified, sector temporally between window_start_time and window_end_time.
        hours_before_sector_time and hours_after_sector_time are ignored if
        window start/end time are set!
    window_end_time: datetime.datetime, default=None
        If specified, sector temporally between window_start_time and window_end_time.
        hours_before_sector_time and hours_after_sector_time are ignored if
        window start/end time are set!

    Returns
    -------
    xarray.Dataset
        sectored dataset containing requested "varnames" sectored spatially and
        temporally as requested.
    """
    from datetime import timedelta

    LOG.debug(
        "Full xarray start/end datetime: %s %s",
        full_xarray.start_datetime,
        full_xarray.end_datetime,
    )
    # numpy.ma.count(full_xarray[varnames[0]].to_masked_array()))

    if area_def is not None:
        if (
            hasattr(area_def, "sector_start_datetime")
            and area_def.sector_start_datetime
        ):
            if not window_start_time:
                # If it is a dynamic sector and we did not specify window
                # start/end time, sector temporally to make sure we use the
                # appropriate data
                mindt = area_def.sector_start_datetime - timedelta(
                    hours=hours_before_sector_time
                )
                maxdt = area_def.sector_start_datetime + timedelta(
                    hours=hours_after_sector_time
                )
            else:
                # If we requested window start/end time, use it.
                mindt = window_start_time
                maxdt = window_end_time
            time_xarray = sector_xarray_temporal(
                full_xarray, mindt, maxdt, varnames, verbose=verbose
            )
        elif window_start_time:
            # If this is not a dynamic sector, but we requested start/end
            # time, still apply the requested window.
            mindt = window_start_time
            maxdt = window_end_time
            time_xarray = sector_xarray_temporal(
                full_xarray, mindt, maxdt, varnames, verbose=verbose
            )
        else:
            # If it is not a dynamic sector, and we didn't pass window
            # start/end time, just return all of the data, because all
            # we care about is spatial coverage.
            time_xarray = full_xarray.copy()

        sector_xarray = sector_xarray_spatial(
            time_xarray,
            area_def,
            varnames,
            lon_pad,
            lat_pad,
            verbose=verbose,
            drop=drop,
        )
        if (
            sector_xarray is not None
            and "time" in varnames
            and hasattr(area_def, "sector_start_datetime")
            and area_def.sector_start_datetime
        ):
            from geoips.xarray_utils.time import (
                get_min_from_xarray_time,
                get_max_from_xarray_time,
            )

            sector_xarray.attrs["area_definition"] = area_def
            sector_xarray.attrs["start_datetime"] = get_min_from_xarray_time(
                sector_xarray, "time"
            )
            sector_xarray.attrs["end_datetime"] = get_max_from_xarray_time(
                sector_xarray, "time"
            )
        elif sector_xarray is not None:
            sector_xarray.attrs["area_definition"] = area_def
            sector_xarray.attrs["start_datetime"] = full_xarray.start_datetime
            sector_xarray.attrs["end_datetime"] = full_xarray.end_datetime

    elif window_start_time:
        # If the area_def is not specified, but we requested start and end time,
        # apply requested window here.
        mindt = window_start_time
        maxdt = window_end_time
        sector_xarray = sector_xarray_temporal(
            full_xarray, mindt, maxdt, varnames, verbose=verbose
        )
    else:
        # If we did not request temporal or spatial sectoring,
        # just return full xarray
        sector_xarray = full_xarray.copy()

    if sector_xarray is not None:
        if verbose:
            LOG.info(
                "Sectored data start/end datetime: %s %s",
                sector_xarray.start_datetime,
                sector_xarray.end_datetime,
            )
            # numpy.ma.count(full_xarray[varnames[0]].to_masked_array()))

    return sector_xarray


def get_vis_ir_bg(sect_xarray):
    """Find matching vis/ir background for data in sect_xarray."""
    from geoips.data_manipulations.merge import get_matching_files
    from geoips.filenames.base_paths import PATHS as gpaths

    irfnames = get_matching_files(
        sect_xarray.area_definition.area_id,
        [sect_xarray.area_definition.area_id],
        ["himawari-8", "goes-16", "goes-17", "msg-1", "msg-4"],
        ["ahi", "abi", "abi", "seviri", "seviri"],
        [30, 30, 30, 30, 30],
        gpaths["PRECALCULATED_DATA_PATH"],
        sect_xarray.start_datetime,
        "Infrared-Gray_latitude_longitude",
        time_format="%H%M*",
        buffer_mins=75,
        verbose=False,
        single_match=True,
    )
    visfnames = get_matching_files(
        sect_xarray.area_definition.area_id,
        [sect_xarray.area_definition.area_id],
        ["himawari-8", "goes-16", "goes-17", "msg-1", "msg-4"],
        ["ahi", "abi", "abi", "seviri", "seviri"],
        [30, 30, 30, 30, 30],
        gpaths["PRECALCULATED_DATA_PATH"],
        sect_xarray.start_datetime,
        "Visible_latitude_longitude",
        time_format="%H%M*",
        buffer_mins=75,
        verbose=False,
        single_match=True,
    )
    LOG.info("irfnames: %s", irfnames)
    LOG.info("visfnames: %s", visfnames)
    ret_arr = {}
    from geoips.plugins.modules.readers.geoips_netcdf import read_xarray_netcdf

    if irfnames:
        ir_bg = read_xarray_netcdf(irfnames[0])
        ret_arr["IR"] = ir_bg
        ret_arr["METADATA"] = ir_bg[[]]
    if visfnames:
        vis_bg = read_xarray_netcdf(visfnames[0])
        ret_arr["VIS"] = vis_bg
        ret_arr["METADATA"] = vis_bg[[]]
    LOG.info("ret_arr: %s", ret_arr)
    return ret_arr


def sector_xarrays(
    xobjs,
    area_def,
    varlist,
    verbose=False,
    hours_before_sector_time=18,
    hours_after_sector_time=6,
    check_center=True,
    drop=False,
    lon_pad=3,
    lat_pad=0,
    window_start_time=None,
    window_end_time=None,
):
    """Return dict of sectored xarray objects.

    Parameters
    ----------
    xobjs: dict of xarray.Dataset
        The full input xarray Datasets, with dataset names as keys to the dictionary.
    area_def: pyresample.AreaDefinition
        The requested region to sector spatially from the full_xarray.
        If None, no spatial sectoring required.
    varlist: list of str
        List of variable names within xarray Datasets to include in the final
        sectored xarray Datasets
    verbose: bool, default=False
        If True, include log output, otherwise return silently.
    hours_before_sector_time: float, default=18
        For dynamic sectors, number of hours before sector start time
        to include when sectoring temporally.
    hours_after_sector_time: float, default=6
        For dynamic sectors, number of hours after sector end time
        to include when sectoring temporally.
    check_center: bool, default=True
        If True and sector_type is "tc", check that there is coverage in the
        center of the image prior to sectoring. Skip dataset altogether if no
        center coverage.
    drop: bool, default=False
        If True, drop data outside range, temporally and spatially.
        If False, mask outside range.
    lon_pad: float, default=3
        Amount of padding to include around the longitude dimension
    lat_pad: float, default=0
        Amount of padding to include around the latitude dimension
    window_start_time: datetime.datetime, default=None
        If specified, sector temporally between window_start_time and window_end_time.
        hours_before_sector_time and hours_after_sector_time are ignored if
        window start/end time are set!
    window_end_time: datetime.datetime, default=None
        If specified, sector temporally between window_start_time and window_end_time.
        hours_before_sector_time and hours_after_sector_time are ignored if
        window start/end time are set!

    Returns
    -------
    dict of xarray.Dataset
        Dictionary of sectored datasets containing requested "varnames",
        sectored spatially and temporally as requested.
    """
    ret_xobjs = {}
    from geoips.xarray_utils.time import (
        get_min_from_xarray_time,
        get_max_from_xarray_time,
    )

    for key, xobj in xobjs.items():
        LOG.info("SECTORING dataset %s area_def %s", key, area_def.name)
        LOG.info(" requested variables %s", set(varlist))
        LOG.info(" dataset variables %s", set(xobj.variables.keys()))
        LOG.info(" dataset data_vars %s", set(xobj.data_vars))
        LOG.info("  Min time: %s", get_min_from_xarray_time(xobj, "time"))
        LOG.info("  Max time: %s", get_max_from_xarray_time(xobj, "time"))
        # Compile a list of variables that will be used to sector - the current data
        # variable, and we will add in the appropriate latitude and longitude variables
        # (of the same shape as data), and if it exists the appropriately shaped time
        # array.
        # Use data_vars here - since coordinates (lat/lon/time) get added below
        vars_to_interp = list(set(varlist) & set(xobj.data_vars))
        if not vars_to_interp:
            LOG.info("  SKIPPING no required variables in dataset %s", key)
            continue

        from geoips.sector_utils.utils import is_dynamic_sector

        if is_dynamic_sector(area_def):
            LOG.info(
                "  Trying to sector %s with dynamic time %s, data time %s to %s, "
                "%s points",
                area_def.area_id,
                area_def.sector_start_datetime,
                xobj.start_datetime,
                xobj.end_datetime,
                xobj["latitude"].size,
            )
        else:
            LOG.info(
                "  Trying to sector %s, %s points",
                area_def.area_id,
                xobj["latitude"].size,
            )

        vars_to_sect = []
        vars_to_sect += vars_to_interp
        # we have to have 'latitude','longitude" in the full_xarray, and 'time' if we
        # want temporal sectoring.
        # Note if lat/lon/time are included as coordinates, they will NOT show up
        # in data_vars, so must use variables
        if "latitude" in list(xobj.variables.keys()):
            vars_to_sect += ["latitude"]
        if "longitude" in list(xobj.variables.keys()):
            vars_to_sect += ["longitude"]
        if "time" in list(xobj.variables.keys()):
            vars_to_sect += ["time"]

        from geoips.xarray_utils.data import sector_xarray_dataset

        # The list of variables in vars_to_sect must ALL be the same shape
        sect_xarray = sector_xarray_dataset(
            xobj,
            area_def,
            vars_to_sect,
            verbose=verbose,
            hours_before_sector_time=hours_before_sector_time,
            hours_after_sector_time=hours_after_sector_time,
            drop=drop,
            lon_pad=lon_pad,
            lat_pad=lat_pad,
            window_start_time=window_start_time,
            window_end_time=window_end_time,
        )

        # numpy arrays fail if numpy_array is None, and xarrays fail if x_array == None
        if sect_xarray is None:
            if verbose:
                LOG.info("  No coverage - skipping dataset %s", key)
            continue

        # check for any obs  near TC center (with a box of 8deg X 8deg), i.e.,
        # within a  range of 400km)
        from geoips.sector_utils.utils import is_sector_type

        if is_sector_type(area_def, "tc") and check_center:
            from geoips.sector_utils.utils import check_center_coverage

            has_covg, covg_xarray = check_center_coverage(
                sect_xarray,
                area_def,
                varlist=vars_to_sect,
                covg_varlist=vars_to_sect,
                width_degrees=8,
                height_degrees=8,
                verbose=verbose,
            )
            if not has_covg:
                LOG.info(
                    "  SKIPPING DATASET %s, NO COVERAGE IN center box - NOT PROCESSING",
                    key,
                )
                continue

            # If the time within the box is > 50 min, we have two overpasses. ALL PMW
            # sensors are polar orbiters.
            if (covg_xarray.end_datetime - covg_xarray.start_datetime).seconds > 3000:
                LOG.info(
                    "Original sectored xarray contains more than one overpass - "
                    "switching to start/datetime in center"
                )
                sect_xarray.attrs["start_datetime"] = covg_xarray.start_datetime
                sect_xarray.attrs["end_datetime"] = covg_xarray.end_datetime

        sect_xarray.attrs["area_definition"] = (
            area_def  # add name of this sector to sector attribute
        )
        if hasattr(sect_xarray, "time"):

            sect_xarray.attrs["start_datetime"] = get_min_from_xarray_time(
                sect_xarray, "time"
            )
            sect_xarray.attrs["end_datetime"] = get_max_from_xarray_time(
                sect_xarray, "time"
            )
            # Note:  need to test whether above two lines can reselect min and max
            # time_info for this sector

        LOG.debug(
            "  Sectored data start/end datetime: %s %s, %s points from var %s, "
            "all vars %s",
            sect_xarray.start_datetime,
            sect_xarray.end_datetime,
            sect_xarray[vars_to_interp[0]].size,
            vars_to_interp[0],
            vars_to_interp,
        )
        sect_xarray.attrs["sectored"] = True
        ret_xobjs[key] = sect_xarray
        ret_xobjs["METADATA"] = sect_xarray[[]]
        LOG.info("  Min time: %s", get_min_from_xarray_time(sect_xarray, "time"))
        LOG.info("  Max time: %s", get_max_from_xarray_time(sect_xarray, "time"))

    return ret_xobjs


def get_sectored_xarrays(
    xobjs, area_def, varlist, get_bg_xarrays=False, check_center=True, drop=False
):
    """Get all xarray objects sectored to area_def.

    Return primary dataset, as well as VIS/IR overlay datasets.
    """
    # All datasets will be of the same data type
    sect_xarrays = sector_xarrays(
        xobjs, area_def, varlist, check_center=check_center, drop=drop
    )
    if get_bg_xarrays and sect_xarrays:
        bg_xarrays = get_vis_ir_bg(sect_xarrays["METADATA"])
        sect_xarrays += sector_xarrays(
            bg_xarrays,
            area_def,
            ["Infrared-Gray", "Visible"],
            check_center=check_center,
            drop=drop,
        )
    else:
        LOG.info(
            "SKIPPING BACKGROUNDS, no coverage for %s %s",
            xobjs[0].source_name,
            area_def.name,
        )
    return sect_xarrays


def combine_preproc_xarrays_with_alg_xarray(dict_of_xarrays, alg_xarray, rgb_var=None):
    """Combine preprocessed xarrays with an xarray output from an algorithm.

    The dimensions of the preprocessed xarrays must match the algorithm xarray.

    Parameters
    ----------
    dict_of_xarrays : dict
        Dictionary of xarray objects (e.g. output from geoips_netcdf reader).
    alg_xarray : xarray dataset
        xarray dataset returned from algorithm.
    rgb_var : bool
        Specify the product name produced by an algorithm that outputs RGB values.
        If so, there will not be any NaNs to replace and a composite will not be made.
        This function will first set all 0s to NaN during the merge process,
        then replace any remaining NaNs back to 0.

    Returns
    -------
    xarray.Dataset
        Combined xarray dataset of input xarrays datasets
    """
    # Merge xarray objects from newest to oldest
    # Otherwise older swaths will be overlaid on newer
    keys = sorted(dict_of_xarrays)[::-1]
    # Set merged xarray object to alg_xarray, NaNs in this xarray dataset will
    # be replaced with increasingly older pre-processed swaths
    merged = alg_xarray
    if rgb_var:
        from numpy import nan

        merged[rgb_var] = merged[rgb_var].where(merged[rgb_var] != 0, nan)
    for i, key in enumerate(keys):
        if key == "METADATA":
            continue
        preproc = dict_of_xarrays[key]
        if rgb_var:
            preproc[rgb_var] = preproc[rgb_var].where(preproc[rgb_var] != 0, nan)
        # # Replace NaNs in new dataset with composited xarray
        # merged = new_dset.combine_first(merged)
        merged = merged.combine_first(preproc)
    if rgb_var:
        merged[rgb_var] = merged[rgb_var].fillna(0)
    return merged



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Ancillary module containing test dataset information.

Dictionary mapping of GeoIPS Test Datasets.

Mapping goes {"test_dataset_name": "test_dataset_url"}
"""

from os import environ

import yaml

interface = None  # denotes that this is not a plugin module

urls = yaml.safe_load(
    open(f"{environ['GEOIPS_PACKAGES_DIR']}/geoips/setup/test-data-urls.yaml", "r")
)

test_dataset_dict = urls["test_data_urls"]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""SSMIS Binary reader.

This code is converted from geoips v1 into geoipd v2 framework. This new version
of reader is indepent from the GEOIPS system whose environmental parametrs must
be used in V1.  Now, only python functipns are used with geoips framework
and xarray is utilized to process datasets for product applications.

Version Histopry:
     V1: initial code, July 24, 2020, NRL-MRY

Input File
     SSMIS SDR data

Output Fields
     XARRAY onjectives to hold variables
"""
# Python Standard Libraries
from datetime import datetime, timedelta
import logging
from os.path import basename

# Third-Party Libraries
import numpy as np
import pandas as pd
import xarray as xr

LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "ssmis_binary"
source_names = ["ssmis"]

# NOTE: Anytime you see a # NOQA comment, this is for flake8 formatting. Unused
# variables are needed in this for moving through the binary file correctly.


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read SSMIS binary data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    LOG.info("Reading files %s", fnames)

    # checking for right input SSMIS SDR file

    xobjs_list = []
    # NOTE chans, area_def, and self_register NOT implemented.
    for fname in fnames:
        xobjs_list += [read_ssmis_data_file(fname, metadata_only=metadata_only)]

    final_xobjs = append_xarray_dicts(xobjs_list)

    return final_xobjs


def append_xarray_dicts(xobjs_list):
    """Append two dictionaries of xarray objects."""
    final_xobjs = {}
    # Loop through all the datasets in the first
    # xarray object - these should be the same in all
    min_start_dt = xobjs_list[0]["METADATA"].attrs["start_datetime"]
    max_end_dt = xobjs_list[0]["METADATA"].attrs["end_datetime"]
    final_xobjs["METADATA"] = xobjs_list[0]["METADATA"][[]]
    for dsname in xobjs_list[0]:
        xobjs_ds_list = []
        for xobjs in xobjs_list:
            xobjs_ds_list += [xobjs[dsname]]
            curr_start_dt = xobjs[dsname].attrs["start_datetime"]
            curr_end_dt = xobjs[dsname].attrs["end_datetime"]
            if curr_start_dt < min_start_dt:
                min_start_dt = curr_start_dt
            if curr_end_dt > max_end_dt:
                max_end_dt = curr_end_dt
        if dsname != "METADATA":
            final_xobjs[dsname] = xr.concat(xobjs_ds_list, dim="dim_0")
        final_xobjs[dsname].attrs["start_datetime"] = min_start_dt
        final_xobjs[dsname].attrs["end_datetime"] = max_end_dt
    final_xobjs["METADATA"].attrs["start_datetime"] = min_start_dt
    final_xobjs["METADATA"].attrs["end_datetime"] = max_end_dt
    return final_xobjs


def read_ssmis_data_file(fname, metadata_only=False):
    """Read a single SSMIS data file."""
    data_name = basename(fname).split("_")[-1].split(".")[-1]

    if data_name != "raw":
        LOG.warning("Warning: wrong SSMI SDR data type:  data_type=%s", data_name)
        raise

    if "cfnoc" in basename(fname) and "sdris" in basename(fname):
        LOG.info("found a SSMIS SDR file")
    else:
        LOG.info("not a SSMIS SDR file: skip it")
        raise

    # ------------------------  Process of dataset ---------------------------

    f1 = open(fname, "rb")

    # READ HEARDER
    sw_rev = np.fromstring(f1.read(2), dtype=np.dtype("short")).byteswap()[0]  # NOQA
    endian, fileid = np.fromstring(
        f1.read(2), dtype=np.dtype("int8")
    ).byteswap()  # NOQA
    rev = np.fromstring(f1.read(4), dtype=np.dtype("int32")).byteswap()
    year = np.fromstring(f1.read(4), dtype=np.dtype("int32")).byteswap()
    jday = np.fromstring(f1.read(2), dtype=np.dtype("short")).byteswap()
    hour, minu = np.fromstring(f1.read(2), dtype=np.dtype("int8")).byteswap()
    satid, nsdr = np.fromstring(f1.read(4), dtype=np.dtype("short")).byteswap()
    spare1, spare2, spare3 = np.fromstring(  # NOQA
        f1.read(3), dtype=np.dtype("int8")
    ).byteswap()
    proc_stat_flags = np.fromstring(  # NOQA
        f1.read(1), dtype=np.dtype("int8")
    ).byteswap()  # NOQA
    spare4 = np.fromstring(f1.read(4), dtype=np.dtype("int32")).byteswap()  # NOQA
    # Need to set up time to be read in by the metadata (year and jday are arrays)
    time = "%04d%03d%02d%02d" % (year[0], jday[0], hour, minu)  # NOQA
    nbytes = 28  # bytes that have been read in
    # Read scan records at 512-byte boundaries
    nfiller = 512 - (
        nbytes % 512
    )  # skip nfiller bytes so that the scan header will start at the 513th byte of
    # the data records,
    filler_bytes = np.fromstring(
        f1.read(nfiller), dtype=np.dtype("int8")
    ).byteswap()  # NOQA

    # Rev 6A of the SSMIS SDR software changed the scalling of channel 12-16 to 100
    # (it was 10 before this change) effective with orbit rev 12216 for F-16 and
    # thereafter for all future satellites
    rev6a = 1
    # When revs wrapped back to 0, the rev[0] < 12216 was no longer valid.
    # Check that rev < 12216, AND year < 2023 (revs wrapped on 7 March 2023)
    if satid == 1 and rev[0] < 12216 and year < 2023:
        rev6a = 0

    if satid == 1:
        satid = "F16"
    elif satid == 2:
        satid = "F17"
    elif satid == 3:
        satid = "F18"
    elif satid == 4:
        satid = "F19"

    satellite_zenith_angle = 53.1
    sensor_scan_angle = 45.0
    satellite_altitude = 859

    # bad_value = -999

    for nn in range(nsdr):  # loop number of sdr data records
        nbytes = 0

        # SCAN HEADER
        syncword = np.fromstring(f1.read(4), dtype=np.dtype("int32")).byteswap()  # NOQA
        scan_year = np.fromstring(f1.read(4), dtype=np.dtype("int32")).byteswap()
        scan_jday = np.fromstring(f1.read(2), dtype=np.dtype("short")).byteswap()
        scan_hour, scan_minu = np.fromstring(
            f1.read(2), dtype=np.dtype("int8")
        ).byteswap()
        scan = np.fromstring(f1.read(4), dtype=np.dtype("int32")).byteswap()  # NOQA
        nscan_imager, nscan_enviro, nscan_las, nscan_uas = np.fromstring(
            f1.read(4), dtype=np.dtype("int8")
        ).byteswap()
        start_scantime_imager = np.fromstring(
            f1.read(112), dtype=np.dtype("int32")
        ).byteswap()
        scenecounts_imager = np.fromstring(
            f1.read(28), dtype=np.dtype("uint8")
        ).byteswap()
        start_scantime_enviro = np.fromstring(
            f1.read(96), dtype=np.dtype("int32")
        ).byteswap()
        scenecounts_enviro = np.fromstring(
            f1.read(24), dtype=np.dtype("uint8")
        ).byteswap()
        start_scantime_las = np.fromstring(
            f1.read(32), dtype=np.dtype("int32")
        ).byteswap()
        scenecounts_las = np.fromstring(f1.read(8), dtype=np.dtype("uint8")).byteswap()
        start_scantime_uas = np.fromstring(
            f1.read(16), dtype=np.dtype("int32")
        ).byteswap()
        scenecounts_uas = np.fromstring(f1.read(4), dtype=np.dtype("uint8")).byteswap()
        spare = np.fromstring(f1.read(20), dtype=np.dtype("int32")).byteswap()  # NOQA
        nbytes += 360  # total bytes of the scan header
        # nscan0 = scan - 1  # number of scans

        # not use geoips functions for time variables
        yyyyjjjhhmn = "{0:4d}{1:03d}{2:02d}{3:02d}".format(
            scan_year[0], scan_jday[0], scan_hour, scan_minu
        )

        # set up start and end toem of this data

        if nn == 0:
            start_time = yyyyjjjhhmn
        # if nn == nsdr - 1:
        #     end_time = yyyyjjjhhmn

        try:
            # The start end time in the standard file name seems to be a bit more
            # reliable, so try parsing the file name first before falling back
            # on the header information in the binary file

            # Grab only the file name
            bfname = basename(fname)

            # Standard SSMIS file names are the following format:
            # <FILE_ID>_<SATELLITE>_<DATE>_<START>_<END>_<REV>_<EXTENSION>
            split_fname = bfname.split("_")
            date = split_fname[2]
            start = split_fname[3]
            end = split_fname[4]
            start_datetime = datetime.strptime(date + start, "d%Y%m%ds%H%M%S")
            end_datetime = datetime.strptime(date + end, "d%Y%m%de%H%M%S")
            if end_datetime < start_datetime:
                # Rolled over to new day
                end_datetime += timedelta(days=1)
        except ValueError:
            # Could not parse the start/end time from the file name
            # Fall back on using info in header
            start_datetime = datetime.strptime(start_time, "%Y%j%H%M")

            # Estimate end time just for metadata purposes
            # To my knowledge, the file end time is not included in the binary header
            # The following offset was determined by taking the start/end time
            # listed in the file name, then dividing the total number of seconds by the
            # number of sensor data records, which came out to 45.5 seconds
            estimated_seconds_per_sensor_data_record = 45.5
            seconds_offset = estimated_seconds_per_sensor_data_record * nsdr
            end_datetime = start_datetime + timedelta(seconds=seconds_offset)

        try:
            # The start end time in the standard file name seems to be a bit more
            # reliable, so try parsing the file name first before falling back
            # on the header information in the binary file

            # Grab only the file name
            bfname = basename(fname)

            # Standard SSMIS file names are the following format:
            # <FILE_ID>_<SATELLITE>_<DATE>_<START>_<END>_<REV>_<EXTENSION>
            split_fname = bfname.split("_")
            date = split_fname[2]
            start = split_fname[3]
            end = split_fname[4]
            start_datetime = datetime.strptime(date + start, "d%Y%m%ds%H%M%S")
            end_datetime = datetime.strptime(date + end, "d%Y%m%de%H%M%S")
            if end_datetime < start_datetime:
                # Rolled over to new day
                end_datetime += timedelta(days=1)
        except ValueError:
            # Could not parse the start/end time from the file name
            # Fall back on using info in header
            start_datetime = datetime.strptime(start_time, "%Y%j%H%M")

            # Estimate end time just for metadata purposes
            # To my knowledge, the file end time is not included in the binary header
            # The following offset was determined by taking the start/end time
            # listed in the file name, then dividing the total number of seconds by the
            # number of sensor data records, which came out to 45.5 seconds
            estimated_seconds_per_sensor_data_record = 45.5
            seconds_offset = estimated_seconds_per_sensor_data_record * nsdr
            end_datetime = start_datetime + timedelta(seconds=seconds_offset)

        try:
            # The start end time in the standard file name seems to be a bit more
            # reliable, so try parsing the file name first before falling back
            # on the header information in the binary file

            # Grab only the file name
            bfname = basename(fname)

            # Standard SSMIS file names are the following format:
            # <FILE_ID>_<SATELLITE>_<DATE>_<START>_<END>_<REV>_<EXTENSION>
            split_fname = bfname.split("_")
            date = split_fname[2]
            start = split_fname[3]
            end = split_fname[4]
            start_datetime = datetime.strptime(date + start, "d%Y%m%ds%H%M%S")
            end_datetime = datetime.strptime(date + end, "d%Y%m%de%H%M%S")
            if end_datetime < start_datetime:
                # Rolled over to new day
                end_datetime += timedelta(days=1)
        except ValueError:
            # Could not parse the start/end time from the file name
            # Fall back on using info in header
            start_datetime = datetime.strptime(start_time, "%Y%j%H%M")

            # Estimate end time just for metadata purposes
            # To my knowledge, the file end time is not included in the binary header
            # The following offset was determined by taking the start/end time
            # listed in the file name, then dividing the total number of seconds by the
            # number of sensor data records, which came out to 45.5 seconds
            estimated_seconds_per_sensor_data_record = 45.5
            seconds_offset = estimated_seconds_per_sensor_data_record * nsdr
            end_datetime = start_datetime + timedelta(seconds=seconds_offset)

        if metadata_only:
            xarray_imager = xr.Dataset()
            xarray_imager.attrs["start_datetime"] = start_datetime
            xarray_imager.attrs["end_datetime"] = end_datetime
            xarray_imager.attrs["source_name"] = "ssmis"
            xarray_imager.attrs["platform_name"] = satid
            xarray_imager.attrs["data_provider"] = "DMSP"
            xarray_imager.attrs["source_file_names"] = [basename(fname)]
            xarray_imager.attrs["satellite_zenith_angle"] = satellite_zenith_angle
            xarray_imager.attrs["sensor_scan_angle"] = sensor_scan_angle
            xarray_imager.attrs["satellite_altitude"] = satellite_altitude
            LOG.info(
                "approximate start_time, end_time= %s, %s",
                xarray_imager.start_datetime,
                xarray_imager.end_datetime,
            )
            return {"METADATA": xarray_imager}

        # -----------------------------------------------------------------------------------------

        #       -------- Apply the GEOIPS framework in XARRAY data frame ----------

        if scenecounts_imager[0] < 0:
            LOG.info("IMAGER is negative")

        # initilization of imager variables

        lt_img = np.zeros((nscan_imager, 180))
        lg_img = np.zeros((nscan_imager, 180))
        ch08 = np.zeros((nscan_imager, 180))
        ch09 = np.zeros((nscan_imager, 180))
        ch10 = np.zeros((nscan_imager, 180))
        ch11 = np.zeros((nscan_imager, 180))
        ch17 = np.zeros((nscan_imager, 180))
        ch18 = np.zeros((nscan_imager, 180))
        surf = np.zeros((nscan_imager, 180))
        rain = np.zeros((nscan_imager, 180))
        time_imager = np.zeros((nscan_imager, 180))

        # IMAGER READ DATA
        for ii in range(nscan_imager):
            if start_scantime_imager[ii] == -999:
                LOG.info("value of imager scan is %s", ii)
                continue
            for jj in range(scenecounts_imager[ii]):
                imager_lat, imager_lon, imager_scene = np.fromstring(
                    f1.read(6), dtype=np.dtype("short")
                ).byteswap()
                imager_surf, imager_rain = np.fromstring(
                    f1.read(2), dtype=np.dtype("int8")
                ).byteswap()
                (
                    imager_ch08,
                    imager_ch09,
                    imager_ch10,
                    imager_ch11,
                    imager_ch17,
                    imager_ch18,
                ) = np.fromstring(f1.read(12), dtype=np.dtype("short")).byteswap()
                nbytes += 20
                # k = 180 * (nscan0 + ii) + jj
                lat = 0.01 * imager_lat
                lon = 0.01 * imager_lon
                try:
                    lt_img[ii][jj] = lat
                    lg_img[ii][jj] = lon
                    ch08[ii][jj] = imager_ch08  # 150    Ghz
                    ch09[ii][jj] = imager_ch09  # 183+-7
                    ch10[ii][jj] = imager_ch10  # 183+-3
                    ch11[ii][jj] = imager_ch11  # 183+-1
                    ch17[ii][jj] = imager_ch17  # 91V
                    ch18[ii][jj] = imager_ch18  # 91H
                    surf[ii][jj] = imager_surf
                    rain[ii][jj] = imager_rain
                except ValueError:
                    LOG.info("Failed setting arrays in scan_imager")

        time_imager[:][
            :
        ] = yyyyjjjhhmn  # set same time for this data record and must use the
        # datetime64 format

        # catenation of data records
        if nn == 0:
            lt_img_pre = lt_img
            lg_img_pre = lg_img
            ch08_pre = ch08
            ch09_pre = ch09
            ch10_pre = ch10
            ch11_pre = ch11
            ch17_pre = ch17
            ch18_pre = ch18
            surf_pre = surf
            rain_pre = rain
            time_imager_pre = time_imager
        else:
            lt_img = np.vstack((lt_img_pre, lt_img))
            lg_img = np.vstack((lg_img_pre, lg_img))
            ch08 = np.vstack((ch08_pre, ch08))
            ch09 = np.vstack((ch09_pre, ch09))
            ch10 = np.vstack((ch10_pre, ch10))
            ch11 = np.vstack((ch11_pre, ch11))
            ch17 = np.vstack((ch17_pre, ch17))
            ch18 = np.vstack((ch18_pre, ch18))
            surf = np.vstack((surf_pre, surf))
            rain = np.vstack((rain_pre, rain))
            time_imager = np.vstack((time_imager_pre, time_imager))

        # reset pre_catinated arrays
        lt_img_pre = lt_img
        lg_img_pre = lg_img
        ch08_pre = ch08
        ch09_pre = ch09
        ch10_pre = ch10
        ch11_pre = ch11
        ch17_pre = ch17
        ch18_pre = ch18
        surf_pre = surf
        rain_pre = rain
        time_imager_pre = time_imager

        # -----------------------------------------------------------------------------------------
        #
        #       for ENVIRONMENTAL Variables

        if scenecounts_enviro[0] < 0:
            LOG.info("ENVIRO is negative")

        # initilization of imager variables

        lt_env = np.zeros((nscan_enviro, 90))
        lg_env = np.zeros((nscan_enviro, 90))
        ch12 = np.zeros((nscan_enviro, 90))
        ch13 = np.zeros((nscan_enviro, 90))
        ch14 = np.zeros((nscan_enviro, 90))
        ch15 = np.zeros((nscan_enviro, 90))
        ch16 = np.zeros((nscan_enviro, 90))
        ch15_5x5 = np.zeros((nscan_enviro, 90))
        ch16_5x5 = np.zeros((nscan_enviro, 90))
        ch17_5x5 = np.zeros((nscan_enviro, 90))
        ch18_5x5 = np.zeros((nscan_enviro, 90))
        ch17_5x4 = np.zeros((nscan_enviro, 90))
        ch18_5x4 = np.zeros((nscan_enviro, 90))
        time_enviro = np.zeros((nscan_enviro, 90))

        # ENVIRO READ DATA
        for ii in range(nscan_enviro):
            if ii % 2 == 0:  # for odd scan numbers
                if start_scantime_enviro[ii] == -999:
                    LOG.info("value of enviro odd scan is %s", ii)
                    continue
                for jj in range(scenecounts_enviro[ii]):
                    enviroodd_lat, enviroodd_lon, enviroodd_scene = np.fromstring(
                        f1.read(6), dtype=np.dtype("short")
                    ).byteswap()
                    enviroodd_seaice, enviroodd_surf = np.fromstring(
                        f1.read(2), dtype=np.dtype("int8")
                    ).byteswap()
                    (
                        enviroodd_ch12,
                        enviroodd_ch13,
                        enviroodd_ch14,
                        enviroodd_ch15,
                        enviroodd_ch16,
                        enviroodd_ch15_5x5,
                        enviroodd_ch16_5x5,
                        enviroodd_ch17_5x5,
                        enviroodd_ch18_5x5,
                        enviroodd_ch17_5x4,
                        enviroodd_ch18_5x4,
                    ) = np.fromstring(f1.read(22), dtype=np.dtype("short")).byteswap()
                    enviroodd_rain1, enviroodd_rain2 = np.fromstring(
                        f1.read(2), dtype=np.dtype("int8")
                    ).byteswap()
                    edr_bitflags = np.fromstring(  # NOQA
                        f1.read(4), dtype=np.dtype("int32")
                    ).byteswap()
                    nbytes += 36
                    lat = 0.01 * enviroodd_lat
                    lon = 0.01 * enviroodd_lon
                    lt_env[ii][jj] = lat
                    lg_env[ii][jj] = lon
                    if rev6a == 1:
                        ch12[ii][jj] = enviroodd_ch12  # 19H
                        ch13[ii][jj] = enviroodd_ch13  # 19V
                        ch14[ii][jj] = enviroodd_ch14  # 22V
                        ch15[ii][jj] = enviroodd_ch15  # 37H
                        ch16[ii][jj] = enviroodd_ch16  # 37V
                        ch15_5x5[ii][jj] = enviroodd_ch15_5x5
                        ch16_5x5[ii][jj] = enviroodd_ch16_5x5
                        ch17_5x5[ii][jj] = enviroodd_ch17_5x5
                        ch18_5x5[ii][jj] = enviroodd_ch18_5x5
                        ch17_5x4[ii][jj] = enviroodd_ch17_5x4
                        ch18_5x4[ii][jj] = enviroodd_ch18_5x4
                    else:
                        ch12[ii][jj] = 10 * enviroodd_ch12
                        ch13[ii][jj] = 10 * enviroodd_ch13
                        ch14[ii][jj] = 10 * enviroodd_ch14
                        ch15[ii][jj] = 10 * enviroodd_ch15
                        ch16[ii][jj] = 10 * enviroodd_ch16
                        ch15_5x5[ii][jj] = 10 * enviroodd_ch15_5x5
                        ch16_5x5[ii][jj] = 10 * enviroodd_ch16_5x5
                        ch17_5x5[ii][jj] = 10 * enviroodd_ch17_5x5
                        ch18_5x5[ii][jj] = 10 * enviroodd_ch18_5x5
                        ch17_5x4[ii][jj] = 10 * enviroodd_ch17_5x4
                        ch18_5x4[ii][jj] = 10 * enviroodd_ch18_5x4

            if (
                ii % 2 == 1
            ):  # for even scan numbers (ch15_5x5, ch16_5x5, ch17_5x5, ch18_5x5,
                # ch17_5x4, ch18_5x4 ??? need a review)
                if start_scantime_enviro[ii] == -999:
                    LOG.info("value of enviro even scan is %s", ii)
                    continue
                for jj in range(scenecounts_enviro[ii]):
                    enviroeven_lat, enviroeven_lon, enviroeven_scene = np.fromstring(
                        f1.read(6), dtype=np.dtype("short")
                    ).byteswap()
                    enviroeven_seaice, enviroeven_surf = np.fromstring(
                        f1.read(2), dtype=np.dtype("int8")
                    ).byteswap()
                    (
                        enviroeven_ch12,
                        enviroeven_ch13,
                        enviroeven_ch14,
                        enviroeven_ch15,
                        enviroeven_ch16,
                    ) = np.fromstring(f1.read(10), dtype=np.dtype("short")).byteswap()
                    nbytes += 18
                    lat = 0.01 * enviroeven_lat
                    lon = 0.01 * enviroeven_lon
                    lt_env[ii][jj] = lat
                    lg_env[ii][jj] = lon
                    if rev6a == 1:
                        ch12[ii][jj] = enviroeven_ch12
                        ch13[ii][jj] = enviroeven_ch13
                        ch14[ii][jj] = enviroeven_ch14
                        ch15[ii][jj] = enviroeven_ch15
                        ch16[ii][jj] = enviroeven_ch16
                    else:
                        ch12[ii][jj] = 10 * enviroeven_ch12
                        ch13[ii][jj] = 10 * enviroeven_ch13
                        ch14[ii][jj] = 10 * enviroeven_ch14
                        ch15[ii][jj] = 10 * enviroeven_ch15
                        ch16[ii][jj] = 10 * enviroeven_ch16

        time_enviro[:][:] = yyyyjjjhhmn

        # catenation of data records
        if nn == 0:
            lt_env_pre = lt_env
            lg_env_pre = lg_env
            ch12_pre = ch12
            ch13_pre = ch13
            ch14_pre = ch14
            ch15_pre = ch15
            ch16_pre = ch16
            time_enviro_pre = time_enviro

            ch15_5x5_pre = ch15_5x5
            ch16_5x5_pre = ch16_5x5
            ch17_5x5_pre = ch17_5x5
            ch18_5x5_pre = ch18_5x5
            ch17_5x4_pre = ch17_5x4
            ch18_5x4_pre = ch18_5x4
        else:
            lt_env = np.vstack((lt_env_pre, lt_env))
            lg_env = np.vstack((lg_env_pre, lg_env))
            ch12 = np.vstack((ch12_pre, ch12))
            ch13 = np.vstack((ch13_pre, ch13))
            ch14 = np.vstack((ch14_pre, ch14))
            ch15 = np.vstack((ch15_pre, ch15))
            ch16 = np.vstack((ch16_pre, ch16))
            time_enviro = np.vstack((time_enviro_pre, time_enviro))

            ch15_5x5 = np.vstack((ch15_5x5_pre, ch15_5x5))
            ch16_5x5 = np.vstack((ch16_5x5_pre, ch16_5x5))
            ch17_5x5 = np.vstack((ch17_5x5_pre, ch17_5x5))
            ch18_5x5 = np.vstack((ch18_5x5_pre, ch18_5x5))
            ch17_5x4 = np.vstack((ch17_5x4_pre, ch17_5x4))
            ch18_5x4 = np.vstack((ch18_5x4_pre, ch18_5x4))

        # reset pre_catinated arrays
        lt_env_pre = lt_env
        lg_env_pre = lg_env
        ch12_pre = ch12
        ch13_pre = ch13
        ch14_pre = ch14
        ch15_pre = ch15
        ch16_pre = ch16
        time_enviro_pre = time_enviro

        ch15_5x5_pre = ch15_5x5
        ch16_5x5_pre = ch16_5x5
        ch17_5x5_pre = ch17_5x5
        ch18_5x5_pre = ch18_5x5
        ch17_5x4_pre = ch17_5x4
        ch18_5x4_pre = ch18_5x4

        # -----------------------------------------------------------------------------------------
        #           Process LAS data record

        if scenecounts_las[0] < 0:
            LOG.info("LAS is negative")

        lt_las = np.zeros((nscan_las, 60))
        lg_las = np.zeros((nscan_las, 60))
        ch01_3x3 = np.zeros((nscan_las, 60))
        ch02_3x3 = np.zeros((nscan_las, 60))
        ch03_3x3 = np.zeros((nscan_las, 60))
        ch04_3x3 = np.zeros((nscan_las, 60))
        ch05_3x3 = np.zeros((nscan_las, 60))
        ch06_3x3 = np.zeros((nscan_las, 60))
        ch07_3x3 = np.zeros((nscan_las, 60))
        ch08_5x5 = np.zeros((nscan_las, 60))
        ch09_5x5 = np.zeros((nscan_las, 60))
        ch10_5x5 = np.zeros((nscan_las, 60))
        ch11_5x5 = np.zeros((nscan_las, 60))
        ch18_5x5_las = np.zeros((nscan_las, 60))
        ch24_3x3 = np.zeros((nscan_las, 60))
        surf_las = np.zeros((nscan_las, 60))
        time_las = np.zeros((nscan_las, 60))
        height_1000mb = np.zeros((nscan_las, 60))

        # LAS READ DATA
        for ii in range(nscan_las):
            if start_scantime_las[ii] == -999:
                LOG.info("value of las scan is %s", ii)
                continue
            for jj in range(scenecounts_las[ii]):
                try:
                    (
                        las_lat,
                        las_lon,
                        las_ch01_3x3,
                        las_ch02_3x3,
                        las_ch03_3x3,
                        las_ch04_3x3,
                        las_ch05_3x3,
                        las_ch06_3x3,
                        las_ch07_3x3,
                        las_ch08_5x5,
                        las_ch09_5x5,
                        las_ch10_5x5,
                        las_ch11_5x5,
                        las_ch18_5x5,
                        las_ch24_3x3,
                        las_height_1000mb,
                        las_surf,
                    ) = np.fromstring(f1.read(34), dtype=np.dtype("short")).byteswap()
                    las_tqflag, las_hqflag = np.fromstring(
                        f1.read(2), dtype=np.dtype("int8")
                    ).byteswap()
                    las_terrain, las_scene = np.fromstring(
                        f1.read(4), dtype=np.dtype("short")
                    ).byteswap()
                except Exception as resp:
                    LOG.info(
                        "Poorly formatted scene # %s, skipping. resp: %s",
                        str(jj),
                        str(resp),
                    )
                    continue

                lat = 0.01 * las_lat
                lon = 0.01 * las_lon
                nbytes += 40
                lt_las[ii][jj] = lat
                lg_las[ii][jj] = lon
                ch01_3x3[ii][jj] = las_ch01_3x3  # 50.3 V
                ch02_3x3[ii][jj] = las_ch02_3x3  # 52.8 V
                ch03_3x3[ii][jj] = las_ch03_3x3  # 53.60V
                ch04_3x3[ii][jj] = las_ch04_3x3  # 54.4 V
                ch05_3x3[ii][jj] = las_ch05_3x3  # 55.5 V
                ch06_3x3[ii][jj] = las_ch06_3x3  # 57.3 RCP
                ch07_3x3[ii][jj] = las_ch07_3x3  # 59.4 RCP
                ch08_5x5[ii][jj] = las_ch08_5x5  # 150 H
                ch09_5x5[ii][jj] = las_ch09_5x5  # 183.31+-7 H
                ch10_5x5[ii][jj] = las_ch10_5x5  # 183.31+-3 H
                ch11_5x5[ii][jj] = las_ch11_5x5  # 183.31+-1 H
                ch18_5x5[ii][jj] = las_ch18_5x5  # 91 H
                ch24_3x3[ii][jj] = las_ch24_3x3  # 60.79+-36+-0.05 RCP
                surf_las[ii][jj] = las_surf
                height_1000mb[ii][jj] = las_height_1000mb

        time_las[:][:] = yyyyjjjhhmn

        # catenation of data records
        if nn == 0:
            lt_las_pre = lt_las
            lg_las_pre = lg_las
            ch01_3x3_pre = ch01_3x3
            ch02_3x3_pre = ch02_3x3
            ch03_3x3_pre = ch03_3x3
            ch04_3x3_pre = ch04_3x3
            ch05_3x3_pre = ch05_3x3
            ch06_3x3_pre = ch06_3x3
            ch07_3x3_pre = ch07_3x3
            ch08_5x5_pre = ch08_5x5
            ch09_5x5_pre = ch09_5x5
            ch10_5x5_pre = ch10_5x5
            ch11_5x5_pre = ch11_5x5
            ch18_5x5_las_pre = ch18_5x5_las
            ch24_3x3_pre = ch24_3x3
            surf_las_pre = surf_las
            height_1000mb_pre = height_1000mb
            time_las_pre = time_las
        else:
            lt_las = np.vstack((lt_las_pre, lt_las))
            lg_las = np.vstack((lg_las_pre, lg_las))
            ch01_3x3 = np.vstack((ch01_3x3_pre, ch01_3x3))
            ch02_3x3 = np.vstack((ch02_3x3_pre, ch02_3x3))
            ch03_3x3 = np.vstack((ch03_3x3_pre, ch03_3x3))
            ch04_3x3 = np.vstack((ch04_3x3_pre, ch04_3x3))
            ch05_3x3 = np.vstack((ch05_3x3_pre, ch05_3x3))
            ch06_3x3 = np.vstack((ch06_3x3_pre, ch06_3x3))
            ch07_3x3 = np.vstack((ch07_3x3_pre, ch07_3x3))
            ch08_5x5 = np.vstack((ch08_5x5_pre, ch08_5x5))
            ch09_5x5 = np.vstack((ch09_5x5_pre, ch09_5x5))
            ch10_5x5 = np.vstack((ch10_5x5_pre, ch10_5x5))
            ch11_5x5 = np.vstack((ch11_5x5_pre, ch11_5x5))
            ch18_5x5_las = np.vstack((ch18_5x5_las_pre, ch18_5x5_las))
            ch24_3x3 = np.vstack((ch24_3x3_pre, ch24_3x3))
            surf_las = np.vstack((surf_las_pre, surf_las))
            height_1000mb = np.vstack((height_1000mb_pre, height_1000mb))
            time_las = np.vstack((time_las_pre, time_las))

        # reset pre_catinated arrays
        lt_las_pre = lt_las
        lg_las_pre = lg_las
        ch01_3x3_pre = ch01_3x3
        ch02_3x3_pre = ch02_3x3
        ch03_3x3_pre = ch03_3x3
        ch04_3x3_pre = ch04_3x3
        ch05_3x3_pre = ch05_3x3
        ch06_3x3_pre = ch06_3x3
        ch07_3x3_pre = ch07_3x3
        ch08_5x5_pre = ch08_5x5
        ch09_5x5_pre = ch09_5x5
        ch10_5x5_pre = ch10_5x5
        ch11_5x5_pre = ch11_5x5
        ch18_5x5_las_pre = ch18_5x5_las
        ch24_3x3_pre = ch24_3x3
        surf_las_pre = surf_las
        height_1000mb_pre = height_1000mb
        time_las_pre = time_las

        # ---------------------------------------------------------------------------------
        #          Process UAS data record

        if scenecounts_uas[0] < 0:
            LOG.info("UAS is negative")

        lt_uas = np.zeros((nscan_uas, 30))
        lg_uas = np.zeros((nscan_uas, 30))
        ch19_6x6 = np.zeros((nscan_uas, 30))
        ch20_6x6 = np.zeros((nscan_uas, 30))
        ch21_6x6 = np.zeros((nscan_uas, 30))
        ch22_6x6 = np.zeros((nscan_uas, 30))
        ch23_6x6 = np.zeros((nscan_uas, 30))
        ch24_6x6 = np.zeros((nscan_uas, 30))
        scene = np.zeros((nscan_uas, 30))
        tqflag = np.zeros((nscan_uas, 30))
        time_uas = np.zeros((nscan_uas, 30))

        # UAS READ DATA
        for ii in range(nscan_uas):
            if start_scantime_uas[ii] == -999:
                LOG.info("value of uas scan is %s", ii)
                continue
            for jj in range(scenecounts_uas[ii]):
                (
                    uas_lat,
                    uas_lon,
                    uas_ch19_6x6,
                    uas_ch20_6x6,
                    uas_ch21_6x6,
                    uas_ch22_6x6,
                    uas_ch23_6x6,
                    uas_ch24_6x6,
                    uas_scene,
                    uas_tqflag,
                ) = np.fromstring(f1.read(20), dtype=np.dtype("short")).byteswap()
                uas_field, uas_bdotk2 = np.fromstring(
                    f1.read(8), dtype=np.dtype("int32")
                ).byteswap()
                nbytes += 28
                lat = 0.01 * uas_lat
                lon = 0.01 * uas_lon
                lt_uas[ii][jj] = lat
                lg_uas[ii][jj] = lon
                ch19_6x6[ii][jj] = uas_ch19_6x6  # 63.28+-0.28 RCP GHz
                ch20_6x6[ii][jj] = uas_ch20_6x6  # 60.79+-0.36 RCP
                ch21_6x6[ii][jj] = uas_ch21_6x6  # 60.79+-0.36+-0.002 RCP
                ch22_6x6[ii][jj] = uas_ch22_6x6  # 60.79+-0.36+-0.0055 RCP
                ch23_6x6[ii][jj] = uas_ch23_6x6  # 60.79+-0.36+-0.0016 RCP
                ch24_6x6[ii][jj] = uas_ch24_6x6  # 60.79+-0.36+-0.050 RCP
                scene[ii][jj] = uas_scene
                tqflag[ii][jj] = uas_tqflag

        time_uas[:][:] = yyyyjjjhhmn

        # catenation of data records
        if nn == 0:
            lt_uas_pre = lt_uas
            lg_uas_pre = lg_uas
            ch19_6x6_pre = ch19_6x6
            ch20_6x6_pre = ch20_6x6
            ch21_6x6_pre = ch21_6x6
            ch22_6x6_pre = ch22_6x6
            ch23_6x6_pre = ch23_6x6
            ch24_6x6_pre = ch24_6x6
            scene_pre = scene
            tqflag_pre = tqflag
            time_uas_pre = time_uas
        else:
            lt_uas = np.vstack((lt_uas_pre, lt_uas))
            lg_uas = np.vstack((lg_uas_pre, lg_uas))
            ch19_6x6 = np.vstack((ch19_6x6_pre, ch19_6x6))
            ch20_6x6 = np.vstack((ch20_6x6_pre, ch20_6x6))
            ch21_6x6 = np.vstack((ch21_6x6_pre, ch21_6x6))
            ch22_6x6 = np.vstack((ch22_6x6_pre, ch22_6x6))
            ch23_6x6 = np.vstack((ch23_6x6_pre, ch23_6x6))
            ch24_6x6 = np.vstack((ch24_6x6_pre, ch24_6x6))
            scene = np.vstack((scene_pre, scene))
            tqflag = np.vstack((tqflag_pre, tqflag))
            time_uas = np.vstack((time_uas_pre, time_uas))

        # reset pre_catinated arrays
        lt_uas_pre = lt_uas
        lg_uas_pre = lg_uas
        ch19_6x6_pre = ch19_6x6
        ch20_6x6_pre = ch20_6x6
        ch21_6x6_pre = ch21_6x6
        ch22_6x6_pre = ch22_6x6
        ch23_6x6_pre = ch23_6x6
        ch24_6x6_pre = ch24_6x6
        scene_pre = scene
        tqflag_pre = tqflag
        time_uas_pre = time_uas

        # -------------------------------------------------------------------------------------------------------
        #       fill up space
        nfiller = 512 - (
            nbytes % 512
        )  # nfiller bytes to be skipped so that the next scan header will start at
        # the 513th byte.
        try:
            filler_bytes = np.fromstring(  # NOQA
                f1.read(nfiller), dtype=np.dtype("int8")
            ).byteswap()[0]
        except Exception as resp:
            LOG.info("Poorly formatted filler_bytes, skipping. resp: %s", str(resp))
            continue

    f1.close()

    LOG.info("start_time, end_time= %s, %s", start_datetime, end_datetime)

    # --------------------- Xarray Objects for Processing Datasets------------
    #   conversion of TBs to K
    #             TBs/100 + 273.15   (K)
    #             setup xarray objects
    namelist_imager = [  # NOQA
        "latitude",
        "longitude",
        "H150",
        "H183-7",
        "H183-3",
        "H183-1",
        "V91",
        "H91",
        "sfcType",
        "rain",
        "time",
    ]
    # namelist_enviro = [
    #     "latitude",
    #     "longitude",
    #     "H19",
    #     "V19",
    #     "V22",
    #     "H37",
    #     "V37",
    #     "ch15_5x5",
    #     "ch16_5x5",
    #     "ch17_5x5",
    #     "ch18_5x5",
    #     "ch17_5x4",
    #     "ch18_5x4",
    #     "time",
    # ]
    # namelist_las = [
    #     "latitude",
    #     "longitude",
    #     "ch01_3x3",
    #     "ch02_3x3",
    #     "ch03_3x3",
    #     "ch04_3x3",
    #     "ch05_3x3",
    #     "ch06_3x3",
    #     "ch07_3x3",
    #     "ch08_5x5",
    #     "ch09_5x5",
    #     "ch10_5x5",
    #     "ch11_5x5",
    #     "ch18_5x5_las",
    #     "ch24_3x3",
    #     "height_1000mb",
    #     "surf_las",
    #     "time",
    # ]
    # namelist_uas = [
    #     "latitude",
    #     "longitude",
    #     "ch19_6x6",
    #     "ch20_6x6",
    #     "ch21_6x6",
    #     "ch22_6x6",
    #     "ch23_6x6",
    #     "ch24_6x6",
    #     "scene",
    #     "tqflag",
    #     "time",
    # ]

    # set xarray object for imager variables
    xarray_imager = xr.Dataset()
    xarray_imager["latitude"] = xr.DataArray(lt_img)
    xarray_imager["longitude"] = xr.DataArray(lg_img)
    # xarray_imager['V150']     = xr.DataArray(ch08/100+273.15)
    # xarray_imager['V150'].attrs['channel_number'] = 8
    xarray_imager["H150"] = xr.DataArray(
        ch08 / 100 + 273.15
    )  # ch08 is 150GHz-H, not 150GHz-V
    xarray_imager["H150"].attrs["channel_number"] = 8
    xarray_imager["H183-7"] = xr.DataArray(ch09 / 100 + 273.15)
    xarray_imager["H183-7"].attrs["channel_number"] = 9
    xarray_imager["H183-3"] = xr.DataArray(ch10 / 100 + 273.15)
    xarray_imager["H183-3"].attrs["channel_number"] = 10
    xarray_imager["H183-1"] = xr.DataArray(ch11 / 100 + 273.15)
    xarray_imager["H183-1"].attrs["channel_number"] = 11
    xarray_imager["V91"] = xr.DataArray(ch17 / 100 + 273.15)
    xarray_imager["V91"].attrs["channel_number"] = 17
    xarray_imager["H91"] = xr.DataArray(ch18 / 100 + 273.15)
    xarray_imager["H91"].attrs["channel_number"] = 18
    xarray_imager["sfcType"] = xr.DataArray(surf)
    xarray_imager["rain"] = xr.DataArray(rain)
    xarray_imager["time"] = xr.DataArray(
        pd.DataFrame(time_imager).astype(int).apply(pd.to_datetime, format="%Y%j%H%M")
    )

    # set xarray object for enviro  variables
    xarray_enviro = xr.Dataset()
    xarray_enviro["latitude"] = xr.DataArray(lt_env)
    xarray_enviro["longitude"] = xr.DataArray(lg_env)
    xarray_enviro["H19"] = xr.DataArray(ch12 / 100 + 273.15)
    xarray_enviro["H19"].attrs["channel_number"] = 12
    xarray_enviro["V19"] = xr.DataArray(ch13 / 100 + 273.15)
    xarray_enviro["V19"].attrs["channel_number"] = 13
    xarray_enviro["V22"] = xr.DataArray(ch14 / 100 + 273.15)
    xarray_enviro["V22"].attrs["channel_number"] = 14
    xarray_enviro["H37"] = xr.DataArray(ch15 / 100 + 273.15)
    xarray_enviro["H37"].attrs["channel_number"] = 15
    xarray_enviro["V37"] = xr.DataArray(ch16 / 100 + 273.15)
    xarray_enviro["V37"].attrs["channel_number"] = 16
    xarray_enviro["Ch15_5x5"] = xr.DataArray(ch15_5x5 / 100 + 273.15)
    xarray_enviro["Ch15_5x5"].attrs["channel_number"] = 15
    xarray_enviro["Ch16_5x5"] = xr.DataArray(ch16_5x5 / 100 + 273.15)
    xarray_enviro["Ch16_5x5"].attrs["channel_number"] = 16
    xarray_enviro["Ch17_5x5"] = xr.DataArray(ch17_5x5 / 100 + 273.15)
    xarray_enviro["Ch17_5x5"].attrs["channel_number"] = 17
    xarray_enviro["Ch18_5x5"] = xr.DataArray(ch18_5x5 / 100 + 273.15)
    xarray_enviro["Ch18_5x5"].attrs["channel_number"] = 18
    xarray_enviro["Ch17_5x4"] = xr.DataArray(ch17_5x4 / 100 + 273.15)
    xarray_enviro["Ch17_5x4"].attrs["channel_number"] = 17
    xarray_enviro["Ch18_5x4"] = xr.DataArray(ch18_5x4 / 100 + 273.15)
    xarray_enviro["Ch18_5x4"].attrs["channel_number"] = 18
    xarray_enviro["time"] = xr.DataArray(
        pd.DataFrame(time_enviro).astype(int).apply(pd.to_datetime, format="%Y%j%H%M")
    )

    # set xarray object for LAS  variables
    xarray_las = xr.Dataset()
    xarray_las["latitude"] = xr.DataArray(lt_las)
    xarray_las["longitude"] = xr.DataArray(lg_las)
    xarray_las["Ch01_3x3"] = xr.DataArray(ch01_3x3 / 100 + 273.15)
    xarray_las["Ch01_3x3"].attrs["channel_number"] = 1
    xarray_las["Ch02_3x3"] = xr.DataArray(ch02_3x3 / 100 + 273.15)
    xarray_las["Ch02_3x3"].attrs["channel_number"] = 2
    xarray_las["Ch03_3x3"] = xr.DataArray(ch03_3x3 / 100 + 273.15)
    xarray_las["Ch03_3x3"].attrs["channel_number"] = 3
    xarray_las["Ch04_3x3"] = xr.DataArray(ch04_3x3 / 100 + 273.15)
    xarray_las["Ch04_3x3"].attrs["channel_number"] = 4
    xarray_las["Ch05_3x3"] = xr.DataArray(ch05_3x3 / 100 + 273.15)
    xarray_las["Ch05_3x3"].attrs["channel_number"] = 5
    xarray_las["Ch06_3x3"] = xr.DataArray(ch06_3x3 / 100 + 273.15)
    xarray_las["Ch06_3x3"].attrs["channel_number"] = 6
    xarray_las["Ch07_3x3"] = xr.DataArray(ch07_3x3 / 100 + 273.15)
    xarray_las["Ch07_3x3"].attrs["channel_number"] = 7
    xarray_las["Ch08_5x5"] = xr.DataArray(ch08_5x5 / 100 + 273.15)
    xarray_las["Ch08_5x5"].attrs["channel_number"] = 8
    xarray_las["Ch09_5x5"] = xr.DataArray(ch09_5x5 / 100 + 273.15)
    xarray_las["Ch09_5x5"].attrs["channel_number"] = 9
    xarray_las["Ch10_5x5"] = xr.DataArray(ch10_5x5 / 100 + 273.15)
    xarray_las["Ch10_5x5"].attrs["channel_number"] = 10
    xarray_las["Ch11_5x5"] = xr.DataArray(ch11_5x5 / 100 + 273.15)
    xarray_las["Ch11_5x5"].attrs["channel_number"] = 11
    xarray_las["Ch18_5x5_las"] = xr.DataArray(ch18_5x5_las / 100 + 273.15)
    xarray_las["Ch18_5x5_las"].attrs["channel_number"] = 18
    xarray_las["Ch24_3x3"] = xr.DataArray(ch24_3x3 / 100 + 273.15)
    xarray_las["Ch24_3x3"].attrs["channel_number"] = 24
    xarray_las["Height_1000mb"] = xr.DataArray(height_1000mb)
    xarray_las["Surf_las"] = xr.DataArray(surf_las)
    xarray_las["time"] = xr.DataArray(
        pd.DataFrame(time_las).astype(int).apply(pd.to_datetime, format="%Y%j%H%M")
    )

    # set xarray object for UAS  variables
    xarray_uas = xr.Dataset()
    xarray_uas["latitude"] = xr.DataArray(lt_uas)
    xarray_uas["longitude"] = xr.DataArray(lg_uas)
    xarray_uas["Ch19_6x6"] = xr.DataArray(ch19_6x6 / 100 + 273.15)
    xarray_uas["Ch19_6x6"].attrs["channel_number"] = 19
    xarray_uas["Ch20_6x6"] = xr.DataArray(ch20_6x6 / 100 + 273.15)
    xarray_uas["Ch20_6x6"].attrs["channel_number"] = 20
    xarray_uas["Ch21_6x6"] = xr.DataArray(ch21_6x6 / 100 + 273.15)
    xarray_uas["Ch21_6x6"].attrs["channel_number"] = 21
    xarray_uas["Ch22_6x6"] = xr.DataArray(ch22_6x6 / 100 + 273.15)
    xarray_uas["Ch22_6x6"].attrs["channel_number"] = 22
    xarray_uas["Ch23_6x6"] = xr.DataArray(ch23_6x6 / 100 + 273.15)
    xarray_uas["Ch23_6x6"].attrs["channel_number"] = 23
    xarray_uas["Ch24_6x6"] = xr.DataArray(ch24_6x6 / 100 + 273.15)
    xarray_uas["Ch24_6x6"].attrs["channel_number"] = 24
    xarray_uas["Scene"] = xr.DataArray(scene)
    xarray_uas["tqFlag"] = xr.DataArray(tqflag)
    xarray_uas["time"] = xr.DataArray(
        pd.DataFrame(time_uas).astype(int).apply(pd.to_datetime, format="%Y%j%H%M")
    )

    # Setup attributes

    # for Imager
    xarray_imager.attrs["start_datetime"] = start_datetime
    xarray_imager.attrs["end_datetime"] = end_datetime
    xarray_imager.attrs["source_name"] = "ssmis"
    xarray_imager.attrs["platform_name"] = satid
    xarray_imager.attrs["data_provider"] = "DMSP"
    xarray_imager.attrs["source_file_names"] = [basename(fname)]
    xarray_imager.attrs["satellite_zenith_angle"] = satellite_zenith_angle
    xarray_imager.attrs["sensor_scan_angle"] = sensor_scan_angle
    xarray_imager.attrs["satellite_altitude"] = satellite_altitude

    # MTIFs need to be "prettier" for PMW products, so 2km resolution for final image
    xarray_imager.attrs["sample_distance_km"] = 2
    xarray_imager.attrs["interpolation_radius_of_influence"] = 15000

    # for Enviro
    xarray_enviro.attrs["start_datetime"] = start_datetime
    xarray_enviro.attrs["end_datetime"] = end_datetime
    xarray_enviro.attrs["source_name"] = "ssmis"
    xarray_enviro.attrs["platform_name"] = satid
    xarray_enviro.attrs["data_provider"] = "DMSP"
    xarray_enviro.attrs["source_file_names"] = [basename(fname)]
    xarray_enviro.attrs["satellite_zenith_angle"] = satellite_zenith_angle
    xarray_enviro.attrs["sensor_scan_angle"] = sensor_scan_angle
    xarray_enviro.attrs["satellite_altitude"] = satellite_altitude

    # MTIFs need to be "prettier" for PMW products, so 2km resolution for final image
    xarray_enviro.attrs["sample_distance_km"] = 2
    xarray_enviro.attrs["interpolation_radius_of_influence"] = 50000

    # for LAS
    xarray_las.attrs["start_datetime"] = start_datetime
    xarray_las.attrs["end_datetime"] = end_datetime
    xarray_las.attrs["source_name"] = "ssmis"
    xarray_las.attrs["platform_name"] = satid
    xarray_las.attrs["data_provider"] = "DMSP"
    xarray_las.attrs["source_file_names"] = [basename(fname)]
    xarray_las.attrs["satellite_zenith_angle"] = satellite_zenith_angle
    xarray_las.attrs["sensor_scan_angle"] = sensor_scan_angle
    xarray_las.attrs["satellite_altitude"] = satellite_altitude

    # MTIFs need to be "prettier" for PMW products, so 2km resolution for final image
    xarray_las.attrs["sample_distance_km"] = 2
    xarray_las.attrs["interpolation_radius_of_influence"] = 50000

    # for UAS
    xarray_uas.attrs["start_datetime"] = start_datetime
    xarray_uas.attrs["end_datetime"] = end_datetime
    xarray_uas.attrs["source_name"] = "ssmis"
    xarray_uas.attrs["platform_name"] = satid
    xarray_uas.attrs["data_provider"] = "DMSP"
    xarray_uas.attrs["source_file_names"] = [basename(fname)]
    xarray_uas.attrs["satellite_zenith_angle"] = satellite_zenith_angle
    xarray_uas.attrs["sensor_scan_angle"] = sensor_scan_angle
    xarray_uas.attrs["satellite_altitude"] = satellite_altitude

    # MTIFs need to be "prettier" for PMW products, so 2km resolution for final image
    xarray_uas.attrs["sample_distance_km"] = 2
    xarray_uas.attrs["interpolation_radius_of_influence"] = 50000

    return {
        "IMAGER": xarray_imager,
        "ENVIRO": xarray_enviro,
        "LAS": xarray_las,
        "UAS": xarray_uas,
        "METADATA": xarray_imager[[]],
    }



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""SSMI binary reader.

This SSMI reader is desgined for importing SSMI sdr data files (such as
ssmi_orbital_sdrmi_f15_d20200427_s104500_e123100_r05323_cfnoc.def). This reader
is created to read in TBs at 19 (V,H),22V, 37(V,H) and 85 (V,H) GHz channels.
There are A/B scans for 85 GHz.  The combined A/B scans will be used for
TC imagery products at 85 GHz.

This GEOIPS python code is based on a SSMI SDR reader in C.

Convert SSMI_HIRES_AB for 85GHz and SSMI_LORES for 19-37GHz into xarray for
GEOIPS framework

V1.0:  Initial version, NRL-MRY, May 19, 2020

SSMI input data info::

    pixels per scan:
    LORES=64 for 19, 22, and 37 GHz channels;
    HIRES=128 for 85 GHz channels
        19V[LORES]                                    FOV:   69km x 43km
        19H[LORES]
        22V[LORES]                                           50km x 40km
        37V[LORES]                                           37km x 28km
        37H[LORES]
        85V[HIRES][2]    [][0]: A scans; [][1]: B scans      15km x 13km
        85H[HIRES][2]

           -----header info-----
    int32
        cyr, cmon, cday,       /* file creation date */
        chr, cmin,             /* file creation time */
        scans,                 /* number of scans in file (from DataSeq) */
        scid,                  /* spacecraft ID */
        rev,                   /* nominal rev  */
        bjld, bhr, bmin, bsec, /* begin day of year (julian day), time=hr,min,sec */
        ejld, ehr, emin, esec, /* ending day of year, time */
        ajld, ahr, amin, asec, /* ascending node DOY, time */
        lsat;                  /* logical satellite ID */

       -----scan data-----
    int32 scann;                 /* scan number (from ScanHdr1) */
    int32 bst;                   /* B-scan start time (sec) scaled by 10000 */
    double xtime;                /* bst as seconds since 0z 1 Jan 1987 */
    uint16
        v19[LORES],              /* TBs */
        h19[LORES],
        v22[LORES],
        v37[LORES],
        h37[LORES],
        v85[HIRES][2],
        h85[HIRES][2],
        lon[HIRES][2];           /* longitudes */
    int16 lat[HIRES][2];         /* latitudes */
    char sft[HIRES][2];          /* surface types */
"""
# Python Standard Libraries
from datetime import datetime
import logging
import os

# Third-Party Libraries
import matplotlib
import numpy as np
import pandas as pd
import xarray as xr


matplotlib.use("agg")

LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "ssmi_binary"
source_names = ["ssmi"]

# NOTE: Anytime you see a # NOQA comment, this is for flake8 formatting. Unused
# variables are needed in this for moving through the binary file correctly.


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read SSMI FNMOC Binary Data.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * NOT YET IMPLEMENTED
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    # from IPython import embed as shell
    fname = fnames[0]

    LOG.info("Reading file %s", fname)

    #     check for right nput SSMI SDR data file

    data_name = os.path.basename(fname).split("_")[-1].split(".")[-1]

    if data_name != "def":
        LOG.info("Warning: wrong SSMI SDR data type:  data_type={0}".format(data_name))
        raise

    if "cfnoc" in os.path.basename(fname) and "sdrmi" in os.path.basename(fname):
        LOG.info("found a SSMI SDR file")
    else:
        LOG.info("not a SSMI SDR file: skip it")
        raise IOError("Not an SSMI SDR file: skip it")

    """    ------  Notes  ------
       Read in SSMI SDR files for 19v-h,22v,37v-h,85v-h(A and B scans) in binary data
            Then, transform these variables (TBs) into xarray framework for GEOIPS

       Input Parameters:
           fname (str): input file name.  require a full path
       Returns:
           xarray.Dataset with required Variables and Attributes:
               Variables:
                        LORES Channels:
                          'latitude', 'longitude', '19V', '19H',
                          '22V', '37V','37H','time_scan'
                        HIRES Channels (combined A-B scans):
                          'latitude', 'longitude', '85V', '85H', 'sfcType', 'time_scan'
               Attibutes:
                        'source_name', 'platform_name', 'data_provider',
                        'interpolation_radius_of_influence',
                        'start_datetime', 'end_datetime'
               Optional Attrs:
                        'source_file_names', 'source_file_datetimes'
    """

    #    *********  Decode Formula for SSMI SDR Data  *************
    # V1 evaluates a 1-byte binary value as an unsigned number
    # V2 evaluates a 2-byte binary value as an unsigned number
    # V4 evaluates a 4-byte binary value as an unsigned number
    # VLat gets latitude and converts its range to [-90:+90]
    # VLon gets longitude

    def V1(p):
        return buf[p]

    def V2(p):
        return 256 * buf[p] + buf[p + 1]

    def V4(p):
        return 256 * (256 * (256 * buf[p] + buf[p + 1]) + buf[p + 2]) + buf[p + 3]

    def VLat(p):
        return V2(p) - 9000

    def VLon(p):
        return V2(p)

    # Define a function to extract values of parameters from a scan data block

    def fill_arrays():
        v19 = {}
        h19 = {}
        v22 = {}
        v37 = {}
        h37 = {}

        lat = {}
        lon = {}
        v85 = {}
        h85 = {}
        sft = {}

        LORES = 64
        unit_scale = 100.0  # convert integer TBs into actual values

        for spotL in range(LORES):
            bb = spotL * 52
            spotH = spotL * 2

            v19[spotL] = V2(bb + 10) / unit_scale
            h19[spotL] = V2(bb + 12) / unit_scale
            v22[spotL] = V2(bb + 14) / unit_scale
            v37[spotL] = V2(bb + 16) / unit_scale
            h37[spotL] = V2(bb + 18) / unit_scale

            # Initilize variable for A and B scans [0] - A scan; [1] - B scan

            # for LORES pixel positions
            lat[spotH] = np.zeros(2)
            lon[spotH] = np.zeros(2)
            v85[spotH] = np.zeros(2)
            h85[spotH] = np.zeros(2)
            sft[spotH] = np.zeros(2)  # surface type

            # for HIRES pixel positions (not at LORES positions)
            lat[spotH + 1] = np.zeros(2)
            lon[spotH + 1] = np.zeros(2)
            v85[spotH + 1] = np.zeros(2)
            h85[spotH + 1] = np.zeros(2)
            sft[spotH + 1] = np.zeros(2)

            # for A scan at LORES pixel positions
            lat[spotH][0] = VLat(bb + 6) / unit_scale
            lon[spotH][0] = VLon(bb + 8) / unit_scale
            v85[spotH][0] = V2(bb + 20) / unit_scale
            h85[spotH][0] = V2(bb + 22) / unit_scale
            sft[spotH][0] = V1(bb + 24)

            # for B scan at LORES pixel positions
            lat[spotH][1] = VLat(bb + 26) / unit_scale
            lon[spotH][1] = VLon(bb + 28) / unit_scale
            v85[spotH][1] = V2(bb + 30) / unit_scale
            h85[spotH][1] = V2(bb + 32) / unit_scale
            sft[spotH][1] = V1(bb + 34)

            # for A scan at HIRES pixel positions
            lat[spotH + 1][0] = VLat(bb + 36) / unit_scale
            lon[spotH + 1][0] = VLon(bb + 38) / unit_scale
            v85[spotH + 1][0] = V2(bb + 40) / unit_scale
            h85[spotH + 1][0] = V2(bb + 42) / unit_scale
            sft[spotH + 1][0] = V1(bb + 44)

            # for B scan at HIRES pixel positions
            lat[spotH + 1][1] = VLat(bb + 46) / unit_scale
            lon[spotH + 1][1] = VLon(bb + 48) / unit_scale
            v85[spotH + 1][1] = V2(bb + 50) / unit_scale
            h85[spotH + 1][1] = V2(bb + 52) / unit_scale
            sft[spotH + 1][1] = V1(bb + 54)

        data = {
            "v19": v19,
            "h19": h19,
            "v22": v22,
            "v37": v37,
            "h37": h37,
            "lat": lat,
            "lon": lon,
            "v85": v85,
            "h85": h85,
            "sft": sft,
        }
        return data

    # Main section of processing SSMI SDR data

    # define some parameters
    LORES = 64  # pixels per lo-res scan
    HIRES = 128  # pixels per gi-res scan
    MAXSCANS = 3000  # max lo-res scans per file
    # SCANTIME = 3.798  # approximate A-B scan interval

    # TRUE = 1
    # FALSE = 0
    # BUFSIZE = 4444
    # FRAMESIZE = 12798
    # FILLER = 0xA5
    EOF_LEN = 6

    # Return Codes
    # OK = 0
    # BAD_HDRS = 3
    # BAD_EOF = 4
    BAD_LEN = 5
    # WRITE_ERR = 6
    # END_FILE = 7
    # FATAL_ERR = 8
    # CANT_OPEN = 9

    # Header Info
    blocks = {
        "ProdID": 28,  # Product ID hdr
        "DataSeq": 26,  # DataSeq hdr  (scan line info, total scans))
        "RevHdrDD": 190,  # Rev info
        "ScanHdrDD": 34,  # Scan hdr
        "ScanDD": 370,  #
        "RevHdr": 30,
        "ScanHdr": 12,
        "Scan": 3334,
    }

    f1 = open(fname, "rb")  # open inform ssmi sdr file

    data = {}  # define a dictionary array for parameters (TBs, etc)

    year = {}  # define arrays for scan ime info
    month = {}
    day = {}
    hour = {}
    minute = {}
    second = {}

    #    Process of READ HEARDERs

    #    Product ID Block
    buf = np.frombuffer(f1.read(blocks["ProdID"]), dtype="uint8")
    satid0 = 10 * (V1(18) - 48) + V1(19) - 48
    fcyr = V2(20)  # date of this input file createed
    # fcmon = V1(22)
    # fcday = V1(23)
    # fchr = V1(24)
    # fcmin = V1(25)

    #    Data Sequence Block
    buf = np.frombuffer(f1.read(blocks["DataSeq"]), dtype="uint8")
    # scans = V2(14)  # number of total scans of this orbital file

    #    Data Description Blocks
    buf = np.frombuffer(f1.read(blocks["RevHdrDD"]), dtype="uint8")
    buf = np.frombuffer(f1.read(blocks["ScanHdrDD"]), dtype="uint8")
    buf = np.frombuffer(f1.read(blocks["ScanDD"]), dtype="uint8")

    #     Rev Header Block
    buf = np.frombuffer(f1.read(blocks["RevHdr"]), dtype="uint8")
    # scid = V4(4)  # spcaecraft ID, i.e., 15 for F15
    # rev = V4(8)
    bjld = V2(12)  # start date info: julian day
    bhr = V1(14)  # -                 hour
    bmin = V1(15)
    # bsec = V1(16)
    ejld = V2(17)  # end date info: Julian day
    ehr = V1(19)
    emin = V1(20)
    # esec = V1(21)
    # ajld = V2(22)  # julian day for ascending node
    # ahr = V1(24)
    # amin = V1(25)
    # asec = V1(26)
    # lsat = V1(27)  # logical satellite ID

    # setup year and julian day for this input file
    year_info = str(fcyr)
    jDay_info = str(bjld)

    #   --------  Start processes of reading and extraction of scans  -------------

    scan_read = 0  # initilization of scan count

    while f1:  # loop scans of this file
        # Read Scan Header Block
        while True:
            buf = np.frombuffer(
                f1.read(2), dtype="uint8"
            )  # read in block length from  two-byte words
            if list(buf) == [0, 0] and len(buf) > 0:
                continue
            elif list(buf)[0] != 165:  # not a FILLER, so get the length of this block
                length = V2(0) * 2
            if length != 0:  # find good block with length-bytes data
                break

        length2 = length - 2  # bytes of the rest block

        buf0 = np.frombuffer(
            f1.read(length2), dtype="uint8"
        )  # read length2 bytes of this block
        buf = np.append([0, 0], buf0)  # shift two bytes so buf will have "length" bytes

        if length == BAD_LEN:
            LOG.info("fatal error:  Ban_length")
            raise  # fatal error stop
        elif length == EOF_LEN or length == 0:
            break
        elif length != blocks["ScanHdr"]:
            LOG.info("block length= {0} {1}".format(blocks["ScanHdr"], length))
            continue  # unexpected block length, go to next block

        # extraction of parameters from scan header block
        # scann = V2(4)  # not used      (first scan, i.e., scan header)
        bst = V4(6)  # B-scan start time (sec): second of the day

        # conver time to seconds from beggining of 1987 (do we need this info?)

        scan_hr = bst / 3600  # hour of the scan
        scan_min = (bst - scan_hr * 3600) / 60  # minute of scan
        scan_sec = bst - scan_hr * 3600 - scan_min * 60  # second of the scan

        date_info = datetime.strptime(year_info + jDay_info, "%Y%j")

        scan_yr = date_info.year
        scan_mon = date_info.month
        # scan_day = date_info.day

        # set up time info for each B scan  (will set: A scantime = B scantime later)
        year[scan_read] = scan_yr
        month[scan_read] = scan_mon
        day[scan_read] = scan_hr
        hour[scan_read] = scan_hr
        minute[scan_read] = scan_min
        second[scan_read] = scan_sec

        # read Scan Data Block

        while True:
            buf = np.frombuffer(
                f1.read(2), dtype="uint8"
            )  # read in block length from  two-byte words
            if list(buf) == [0, 0] and len(buf) > 0:
                continue
            elif list(buf)[0] != 165:  # not a FILLER, so get the length of this block
                length = V2(0) * 2
            if length != 0:  # find good block with length-bytes data
                break

        length2 = length - 2  # bytes of the rest block

        buf0 = np.frombuffer(
            f1.read(length2), dtype="uint8"
        )  # read length2 bytes of this block
        buf = np.append([0, 0], buf0)  # shift two bytes so buf will have "length" bytes

        if length == BAD_LEN:
            raise  # fatal error stop
        elif length == EOF_LEN or length == 0:
            break
        elif length != blocks["Scan"]:
            LOG.info("block length= {0} {1}".format(blocks["Scan"], length))
            continue  # unexpected block lengthi, go to next block

        # check of max scans
        if scan_read > MAXSCANS:
            LOG.info("Reached max scans, break!")
            break

        # extract parameters from the scan block
        data[scan_read] = fill_arrays()

        scan_read += 1  # accumulation of scans

    f1.close()  # close input file

    #  -------- Apply the GEOIPS framework in XARRAY data frame ----------

    LOG.info("Making full dataframe")

    # bad_value = -999

    # initilization of variables
    lat_lo = np.zeros((scan_read, 64))  # LORES channels: lat
    lon_lo = np.zeros((scan_read, 64))  # lon
    V19 = np.zeros((scan_read, 64))
    H19 = np.zeros((scan_read, 64))
    V22 = np.zeros((scan_read, 64))
    V37 = np.zeros((scan_read, 64))
    H37 = np.zeros((scan_read, 64))
    time_scan_lo = np.zeros((scan_read, 64))  # same for every pixel of this scan

    lat_hia = np.zeros((scan_read, 128))  # A scan HIRES channels: lat
    lon_hia = np.zeros((scan_read, 128))  # -                      lon
    lat_hib = np.zeros((scan_read, 128))  # B scan HIRES channels: lat
    lon_hib = np.zeros((scan_read, 128))  # -                      lon
    V85a = np.zeros((scan_read, 128))
    V85b = np.zeros((scan_read, 128))
    H85a = np.zeros((scan_read, 128))
    H85b = np.zeros((scan_read, 128))

    lat_ab = np.zeros((scan_read * 2, 128))  # combined A-B scans: lat
    lon_ab = np.zeros((scan_read * 2, 128))  # -                   lon
    V85 = np.zeros((scan_read * 2, 128))
    H85 = np.zeros((scan_read * 2, 128))
    sfcType = np.zeros((scan_read * 2, 128))
    time_scan = np.zeros((scan_read * 2, 128))  # same for every pixel of this scan

    # assignment of data for variables

    for ii in range(scan_read):  # loop of scans
        ii2 = ii * 2  # for combined A-B scans

        for jj in range(LORES):  # loop of pixels per scan for LORES channels
            jj2 = jj * 2
            try:
                lat_lo[ii][jj] = data[ii]["lat"][jj2][0]
                lon_lo[ii][jj] = data[ii]["lon"][jj2][0]
                V19[ii][jj] = data[ii]["v19"][jj]
                H19[ii][jj] = data[ii]["h19"][jj]
                V22[ii][jj] = data[ii]["v22"][jj]
                V37[ii][jj] = data[ii]["v37"][jj]
                H37[ii][jj] = data[ii]["h37"][jj]
                time_scan_lo[ii][jj] = "%04d%03d%02d%02d" % (
                    fcyr,
                    bjld,
                    hour[ii],
                    minute[ii],
                )
            except KeyError:
                LOG.info(
                    "Failed setting arrays in LORES channels {0} {1} {2}".format(
                        ii, jj, jj2
                    )
                )

        for jj in range(HIRES):  # loop of pixels per scan for HIRES channels
            try:
                lat_hia[ii][jj] = data[ii]["lat"][jj][0]  # A scan
                lon_hia[ii][jj] = data[ii]["lon"][jj][0]
                lat_hib[ii][jj] = data[ii]["lat"][jj][1]  # B scan
                lon_hib[ii][jj] = data[ii]["lon"][jj][1]
                V85a[ii][jj] = data[ii]["v85"][jj][0]  # A scan
                V85b[ii][jj] = data[ii]["v85"][jj][1]  # B scan
                H85a[ii][jj] = data[ii]["h85"][jj][0]  # A scan
                H85b[ii][jj] = data[ii]["h85"][jj][1]  # B scan

                # combined A-B scans
                lat_ab[ii2][jj] = lat_hia[ii][jj]  # put A scan values
                lon_ab[ii2][jj] = lon_hia[ii][jj]
                V85[ii2][jj] = V85a[ii][jj]
                H85[ii2][jj] = H85a[ii][jj]
                sfcType[ii2][jj] = data[ii]["sft"][jj][0]
                time_scan[ii2][jj] = "%04d%03d%02d%02d" % (
                    fcyr,
                    bjld,
                    hour[ii],
                    minute[ii],
                )

                lat_ab[ii2 + 1][jj] = lat_hib[ii][jj]  # put B scan values
                lon_ab[ii2 + 1][jj] = lon_hib[ii][jj]
                V85[ii2 + 1][jj] = V85b[ii][jj]
                H85[ii2 + 1][jj] = H85b[ii][jj]
                sfcType[ii2 + 1][jj] = data[ii]["sft"][jj][1]
                time_scan[ii2 + 1][jj] = time_scan[ii2][
                    jj
                ]  # same time for A and B scan
            except KeyError:
                LOG.info("Failed setting arrays in HIRES channels")

    #          ------  setup xarray variables   ------
    # namelist_lores = [
    #     "latitude",
    #     "longitude",
    #     "V19",
    #     "H19",
    #     "V22",
    #     "V37",
    #     "H37",
    #     "time_scan_lo",
    # ]
    # namelist_85ab = ["latitude", "longitude", "V85", "H85", "sfcType", "time"]

    # for LORES channels
    xarray_lores = xr.Dataset()
    xarray_lores["latitude"] = xr.DataArray(lat_lo)
    xarray_lores["longitude"] = xr.DataArray(lon_lo)
    xarray_lores["V19"] = xr.DataArray(V19)
    xarray_lores["H19"] = xr.DataArray(H19)
    xarray_lores["V22"] = xr.DataArray(V22)
    xarray_lores["V37"] = xr.DataArray(V37)
    xarray_lores["H37"] = xr.DataArray(H37)
    xarray_lores["time"] = xr.DataArray(
        pd.DataFrame(time_scan_lo).astype(int).apply(pd.to_datetime, format="%Y%j%H%M")
    )

    # for combined 85GHz A-B channels
    xarray_85ab = xr.Dataset()
    xarray_85ab["latitude"] = xr.DataArray(lat_ab)
    xarray_85ab["longitude"] = xr.DataArray(lon_ab)
    xarray_85ab["V85"] = xr.DataArray(V85)
    xarray_85ab["H85"] = xr.DataArray(H85)
    xarray_85ab["sfcType"] = xr.DataArray(sfcType)
    xarray_85ab["time"] = xr.DataArray(
        pd.DataFrame(time_scan).astype(int).apply(pd.to_datetime, format="%Y%j%H%M")
    )

    # setup attributes

    # ch:       19   22   37  85    GHz
    # FOV_list=['43','40','28',13']  km

    # satID string
    satid = "F" + str(satid0)

    start_time = "%04d%03d%02d%02d" % (fcyr, bjld, bhr, bmin)
    end_time = "%04d%03d%02d%02d" % (fcyr, ejld, ehr, emin)

    # for LORES
    xarray_lores.attrs["start_datetime"] = datetime.strptime(start_time, "%Y%j%H%M")
    xarray_lores.attrs["end_datetime"] = datetime.strptime(end_time, "%Y%j%H%M")
    xarray_lores.attrs["source_name"] = "ssmi"
    xarray_lores.attrs["platform_name"] = satid
    xarray_lores.attrs["data_provider"] = "DMSP"
    xarray_lores.attrs["source_file_names"] = [os.path.basename(fname)]

    # MTIFs need to be "prettier" for PMW products, so 2km resolution for final image
    # xarray_lores.attrs['sample_distance_km'] = 25
    xarray_lores.attrs["sample_distance_km"] = 2
    xarray_lores.attrs["interpolation_radius_of_influence"] = 50000

    # for 85GHz A-B combined
    xarray_85ab.attrs["start_datetime"] = datetime.strptime(start_time, "%Y%j%H%M")
    xarray_85ab.attrs["end_datetime"] = datetime.strptime(end_time, "%Y%j%H%M")
    xarray_85ab.attrs["source_file_datetimes"] = [xarray_85ab.start_datetime]
    xarray_85ab.attrs["source_name"] = "ssmi"
    xarray_85ab.attrs["platform_name"] = satid
    xarray_85ab.attrs["data_provider"] = "DMSP"
    xarray_85ab.attrs["source_file_names"] = [os.path.basename(fname)]

    # MTIFs need to be "prettier" for PMW products, so 2km resolution for final image
    # xarray_85ab.attrs['sample_distance_km'] = 13
    xarray_85ab.attrs["sample_distance_km"] = 2
    xarray_85ab.attrs["interpolation_radius_of_influence"] = 15000

    LOG.info("  Start time %s", start_time)
    LOG.info("  End time %s", end_time)
    LOG.info("  Min lat %s", lat_ab.min())
    LOG.info("  Max lat %s", lat_ab.max())

    return {"HIRES": xarray_85ab, "LORES": xarray_lores, "METADATA": xarray_85ab[[]]}



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Matplotlib-based windbarb annotated image output."""

import logging

import numpy
import cartopy.crs as crs

from geoips.image_utils.mpl_utils import (
    create_figure_and_main_ax_and_mapobj,
    save_image,
)
from geoips.image_utils.colormap_utils import set_matplotlib_colors_standard
from geoips.image_utils.mpl_utils import plot_image, plot_overlays, create_colorbar
from geoips.image_utils.mpl_utils import get_title_string_from_objects, set_title

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "image_overlay"
name = "imagery_windbarbs"


def plot_barbs(
    main_ax, mapobj, mpl_colors_info, formatted_data_dict, barb_color_variable="speed"
):
    """Plot windbarbs on matplotlib figure."""
    # main_ax.extent = area_def.area_extent_ll
    main_ax.set_extent(mapobj.bounds, crs=mapobj)
    # main_ax.extent = mapobj.bounds
    # NOTE this does not work if transform=mapobj.
    # Something about transforming to PlateCarree projection, then
    # reprojecting to mapobj.  I don't fully understand it, but this
    # works beautifully, and transform=mapobj puts all the vectors
    # in the center of the image.
    main_ax.scatter(
        x=formatted_data_dict["lon"].data[formatted_data_dict["rain_inds"]],
        y=formatted_data_dict["lat"].data[formatted_data_dict["rain_inds"]],
        transform=crs.PlateCarree(),
        marker="D",
        color="k",
        s=formatted_data_dict["rain_size"],
        zorder=2,
    )
    main_ax.barbs(
        formatted_data_dict["lon"].data,
        formatted_data_dict["lat"].data,
        formatted_data_dict["u"].data,
        formatted_data_dict["v"].data,
        formatted_data_dict[barb_color_variable].data,
        transform=crs.PlateCarree(),
        pivot="tip",
        rounding=False,
        cmap=mpl_colors_info["cmap"],
        flip_barb=formatted_data_dict["flip_barb"],
        # barb_increments=dict(half=10, full=20, flag=50),
        sizes=formatted_data_dict["sizes_dict"],
        length=formatted_data_dict["barb_length"],
        linewidth=formatted_data_dict["line_width"],
        norm=mpl_colors_info["norm"],
        zorder=1,
    )


def output_clean_windbarbs(
    area_def,
    clean_fnames,
    mpl_colors_info,
    image_datetime,
    formatted_data_dict,
    fig=None,
    main_ax=None,
    mapobj=None,
    barb_color_variable="speed",
):
    """Plot and save "clean" windbarb imagery.

    No background imagery, coastlines, gridlines, titles, etc.

    Returns
    -------
    list of str
        Full paths to all resulting output files.
    """
    LOG.info("Starting clean_fname")
    if fig is None and main_ax is None and mapobj is None:
        # Create matplotlib figure and main axis, where the main image will be plotted
        fig, main_ax, mapobj = create_figure_and_main_ax_and_mapobj(
            area_def.x_size, area_def.y_size, area_def, noborder=True
        )

    plot_barbs(
        main_ax,
        mapobj,
        mpl_colors_info,
        formatted_data_dict,
        barb_color_variable=barb_color_variable,
    )

    success_outputs = []

    if clean_fnames is not None:
        for clean_fname in clean_fnames:
            success_outputs += save_image(
                fig, clean_fname, is_final=False, image_datetime=image_datetime
            )

    return success_outputs


def format_windbarb_data(xarray_obj, product_name):
    """Format windbarb data before plotting."""
    # lat=xarray_obj['latitude'].to_masked_array()
    # lon2=xarray_obj['longitude'].to_masked_array()
    # direction=xarray_obj['wind_dir_deg_met'].to_masked_array()
    # speed=xarray_obj['wind_speed_kts'].to_masked_array()

    # u=speed * numpy.sin((direction+180)*3.1415926/180.0)
    # v=speed * numpy.cos((direction+180)*3.1415926/180.0)

    # u=speed * numpy.sin(direction*3.1415926/180.0)
    # v=speed * numpy.cos(direction*3.1415926/180.0)

    data_cols = xarray_obj[product_name].attrs.get(
        "windbarb_data_columns", ["speed", "direction", "rain_flag"]
    )

    num_product_arrays = 1
    if len(xarray_obj[product_name].shape) == 3:
        num_product_arrays = xarray_obj[product_name].shape[2]

    # This is 2-D, with only one array per variable (speed, direction,
    # rain_flag) - meaning NO ambiguities
    if len(xarray_obj[product_name].shape) == 3 and num_product_arrays == 3:
        speed = xarray_obj[product_name].to_masked_array()[:, :, 0]
        direction = xarray_obj[product_name].to_masked_array()[:, :, 1]
        rain_flag = xarray_obj[product_name].to_masked_array()[:, :, 2]
    # This is 2-D, with FOUR arrays per variable (speed, direction, rain_flag)
    # - meaning 4 ambiguities
    elif len(xarray_obj[product_name].shape) == 3 and num_product_arrays == 12:
        speed = xarray_obj[product_name].to_masked_array()[:, :, 0:4]
        direction = xarray_obj[product_name].to_masked_array()[:, :, 4:8]
        rain_flag = xarray_obj[product_name].to_masked_array()[:, :, 8:12]
    # This is 1-D, with one vector per variable - no ambiguities.
    elif xarray_obj[product_name].ndim == 3:
        speed = xarray_obj[product_name].to_masked_array()[:, :, 0]
        direction = xarray_obj[product_name].to_masked_array()[:, :, 1]
        rain_flag = xarray_obj[product_name].to_masked_array()[:, :, 2]
        if "pressure" in data_cols:
            pressure = xarray_obj[product_name].to_masked_array()[:, :, 3]
    else:
        speed = xarray_obj[product_name].to_masked_array()[:, 0]
        direction = xarray_obj[product_name].to_masked_array()[:, 1]
        rain_flag = xarray_obj[product_name].to_masked_array()[:, 2]
        if "pressure" in data_cols:
            pressure = xarray_obj[product_name].to_masked_array()[:, 3]
    # These should probably be specified in the product dictionary.
    # It will vary per-sensor / data type, these basically only currently work with
    # ASCAT 25 km data.
    # This would also avoid having the product names hard coded in the output
    # module code.

    from geoips.interfaces import products

    source_prod_spec = products.get_plugin(xarray_obj.source_name, product_name)
    prod_plugin = xarray_obj.attrs.get("product_plugin", source_prod_spec)

    try:
        barb_args = prod_plugin["spec"]["windbarb_plotter"]["plugin"]["arguments"]
    except KeyError:
        barb_args = {}

    if barb_args:
        # Thinning the data points to better display the windbards
        thinning = barb_args["thinning"]
        barblength = barb_args["length"]
        linewidth = barb_args["width"]
        sizes_dict = barb_args["sizes_dict"]
        rain_size = barb_args["rain_size"]
    elif product_name == "windbarbs":
        # Thinning the data points to better display the windbards
        thinning = 1  # skip data points
        barblength = 5.0
        linewidth = 1.5
        sizes_dict = dict(height=0.7, spacing=0.3)
        rain_size = 10
    elif product_name == "wind-ambiguities" or "wind-ambiguities" in product_name:
        # Thinning the data points to better display the windbards
        thinning = 1  # skip data points
        barblength = 5  # Length of individual barbs
        linewidth = 2  # Width of individual barbs
        rain_size = 10  # Marker size for rain_flag
        sizes_dict = dict(
            height=0,
            spacing=0,
            width=0,  # flag width, relative to barblength
            emptybarb=0.5,
        )
    else:
        raise ValueError(f"Unknown product {product_name}")

    lat = xarray_obj["latitude"].to_masked_array()
    lon2 = xarray_obj["longitude"].to_masked_array()
    u = speed * numpy.sin((direction + 180) * 3.1415926 / 180.0)
    v = speed * numpy.cos((direction + 180) * 3.1415926 / 180.0)

    # convert longitudes to (-180,180)
    # lon=utils.wrap_longitudes(lon2)
    # Must be 0-360 for barbs
    lon = numpy.ma.where(lon2 < 0, lon2 + 360, lon2)
    thin_slice = [slice(0, None, thinning)] * lat.ndim

    lat2 = lat[tuple(thin_slice)]
    lon2 = lon[tuple(thin_slice)]
    u2 = u[tuple(thin_slice)]
    v2 = v[tuple(thin_slice)]
    speed2 = speed[tuple(thin_slice)]
    rain_flag2 = rain_flag[tuple(thin_slice)]
    if "pressure" in data_cols:
        pressure2 = pressure[tuple(thin_slice)]

    if lat2.min() > 0:
        flip_barb = False
    elif lat2.max() < 0:
        flip_barb = True
    else:
        flip_barb = numpy.ma.where(lat2 > 0, False, True).data
    good_inds = numpy.ma.where(speed2)
    return_dict = {}
    if len(lon2.shape) != len(speed2.shape):
        return_dict["lon"] = lon2[good_inds[0:2]]
    else:
        return_dict["lon"] = lon2[good_inds]
    if len(lat2.shape) != len(speed2.shape):
        return_dict["lat"] = lat2[good_inds[0:2]]
    else:
        return_dict["lat"] = lat2[good_inds]
    if flip_barb is not True and flip_barb is not False:
        if len(flip_barb.shape) != len(speed2.shape):
            return_dict["flip_barb"] = flip_barb[good_inds[0:2]]
        else:
            return_dict["flip_barb"] = flip_barb[good_inds]
    else:
        return_dict["flip_barb"] = flip_barb
    return_dict["u"] = u2[good_inds]
    return_dict["v"] = v2[good_inds]
    return_dict["speed"] = speed2[good_inds]
    return_dict["rain_inds"] = numpy.ma.where(rain_flag2[good_inds])
    return_dict["barb_length"] = barblength
    return_dict["line_width"] = linewidth
    return_dict["sizes_dict"] = sizes_dict
    return_dict["rain_size"] = rain_size
    if "pressure" in data_cols:
        return_dict["pressure"] = pressure2[good_inds]

    return return_dict


def call(
    area_def,
    xarray_obj,
    product_name,
    output_fnames,
    clean_fname=None,
    product_name_title=None,
    mpl_colors_info=None,
    feature_annotator=None,
    gridline_annotator=None,
    product_datatype_title=None,
    bg_data=None,
    bg_mpl_colors_info=None,
    bg_xarray=None,
    bg_product_name_title=None,
    bg_datatype_title=None,
    remove_duplicate_minrange=None,
    title_copyright=None,
    title_formatter=None,
):
    """Plot annotated windbarbs on matplotlib figure."""
    LOG.info("Startig imagery_windbarbs")

    if product_name_title is None:
        product_name_title = product_name

    success_outputs = []

    bkgrnd_clr = None
    frame_clr = None
    # If a feature_annotator plugin was supplied, attempt to get the image background
    # color. Otherwise, just keep it as None.
    if feature_annotator:
        bkgrnd_clr = feature_annotator.get("spec", {}).get("background")
    # If a gridline_annotator plugin was supplied, attempt to get the frame background
    # color. Otherwise, just keep it as None.
    if gridline_annotator:
        frame_clr = gridline_annotator.get("spec", {}).get("background")

    # Plot windbarbs

    formatted_data_dict = format_windbarb_data(xarray_obj, product_name)

    if clean_fname is not None:
        success_outputs += output_clean_windbarbs(
            area_def,
            [clean_fname],
            mpl_colors_info,
            xarray_obj.start_datetime,
            formatted_data_dict,
        )

    if output_fnames is not None:
        LOG.info("Starting output_fnames")

        # Create matplotlib figure and main axis, where the main image will be plotted
        fig, main_ax, mapobj = create_figure_and_main_ax_and_mapobj(
            area_def.x_size,
            area_def.y_size,
            area_def,
            existing_mapobj=None,
            noborder=False,
            frame_clr=frame_clr,
        )

        if bg_data is not None:
            if not bg_mpl_colors_info:
                bg_mpl_colors_info = set_matplotlib_colors_standard(
                    data_range=[bg_data.min(), bg_data.max()],
                    cmap_name="Greys",
                    cbar_label=None,
                    create_colorbar=False,
                )
            # Plot the background data on a map
            plot_image(
                main_ax,
                bg_data,
                mapobj,
                mpl_colors_info=bg_mpl_colors_info,
                bkgrnd_clr=bkgrnd_clr,
            )

        plot_barbs(main_ax, mapobj, mpl_colors_info, formatted_data_dict)

        # Set the title for final image
        title_string = get_title_string_from_objects(
            area_def,
            xarray_obj,
            product_name_title,
            product_datatype_title=product_datatype_title,
            bg_xarray=bg_xarray,
            bg_product_name_title=bg_product_name_title,
            bg_datatype_title=bg_datatype_title,
            title_copyright=title_copyright,
            title_formatter=title_formatter,
        )
        set_title(main_ax, title_string, area_def.y_size)

        if mpl_colors_info["colorbar"] is True:
            # Create the colorbar to match the mpl_colors
            create_colorbar(fig, mpl_colors_info)

        # Plot gridlines and feature overlays
        plot_overlays(
            mapobj,
            main_ax,
            area_def,
            feature_annotator=feature_annotator,
            gridline_annotator=gridline_annotator,
        )

        for annotated_fname in output_fnames:
            # Save the final image
            success_outputs += save_image(
                fig,
                annotated_fname,
                is_final=True,
                image_datetime=xarray_obj.start_datetime,
            )

    return success_outputs



#!/usr/bin/env python

# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""GeoIPS command line interface."""

import pkgutil
from argparse import (
    ArgumentParser,
    ArgumentDefaultsHelpFormatter,
    RawDescriptionHelpFormatter,
)
from geoips import dev, interfaces
from geoips.interfaces.base import BaseInterface
import warnings
import logging

# Always actually raise DeprecationWarnings
# Note this SO answer https://stackoverflow.com/a/20960427
warnings.simplefilter("always", DeprecationWarning)
LOG = logging.getLogger(__name__)

__doc__ = """
GeoIPS

The Geolocated Information Processing System (GeoIPS) is a generalized
processing system, providing a collection of algorithm and product
implementations facilitating consistent and reliable application of specific
products across a variety of sensors and data types.

GeoIPS acts as a toolbox for internal GeoIPS-based product development - all
modules are expected to have simple inputs and outputs (Python numpy or dask
arrays or xarrays, dictionaries, strings, lists), to enable portability and
simplified interfacing between modules.
"""


class RawDescriptionArgumentDefaultsHelpFormatter(
    ArgumentDefaultsHelpFormatter, RawDescriptionHelpFormatter
):
    """Compound formatter class for user-readable help.

    * preserves the raw description formatting
    * adds defaults to helps.
    """

    pass


formclass = RawDescriptionArgumentDefaultsHelpFormatter


def print_table(title, headings, rows):
    """Print a column formatted table.

    Parameters
    ----------
    title : str
        A title for the table
    headings : list of str
        A list of strings to use as column headings
    rows : list of tupl of str
        A list of equal-length tuples
    """
    ncol = len(headings)
    rows = list(rows)
    for row_ind, row in enumerate(rows):
        if len(row) != ncol:
            raise ValueError("Number of elements per row must be equal for all rows")
        rows[row_ind] = [str(val) if val is not None else "" for val in row]

    col_widths = []
    for col_ind, heading in enumerate(headings):
        col_width = len(heading)
        for row in rows:
            col_width = max(col_width, len(row[col_ind]))
        col_widths.append(col_width)

    title_case_title = title.replace("_", " ").title()
    title_width = sum(col_widths) + 3 * (len(col_widths) - 1)
    LOG.info("\n" + title_width * "=")
    LOG.info(f"{title_case_title:^{title_width}}")
    LOG.info(title_width * "=")

    # Create and print the header
    head_str = " | ".join(
        [f"{head:{width}}" for head, width in zip(headings, col_widths)]
    )
    LOG.info(head_str)
    LOG.info(len(head_str) * "-")

    # Create and print the rows
    for row in rows:
        # print(list(zip(row, col_widths)))
        row_str = " | ".join([f"{val:{width}}" for val, width in zip(row, col_widths)])
        LOG.info(row_str)


def get_interface(name):
    """Get interface."""
    try:
        return getattr(interfaces, name)
    except AttributeError:
        try:
            return getattr(dev, name)
        except AttributeError:
            raise AttributeError(f'Interface "{name}" not found')


def add_list_interface_parser(subparsers, name, aliases=None):
    """Add list interface parser."""
    # list interpolators
    interface_parser = subparsers.add_parser(
        name,
        aliases=aliases,
        description=f"List available {name}",
        help=f"List available {name}",
        formatter_class=formclass,
    )
    interface_parser.set_defaults(interface=name)
    return interface_parser


def list_dev_interfaces():
    """Return a list of all developmental interfaces."""
    return [name for _, name, _ in pkgutil.iter_modules(dev.__path__)]


def list_interfaces(dev=False):
    """List interfaces."""
    inter_names = []
    inter_descs = []
    for attr in dir(interfaces):
        attr_val = getattr(interfaces, attr)
        if issubclass(attr_val.__class__, BaseInterface):
            inter_names.append(attr_val.name)
            try:
                inter_descs.append(getattr(attr_val, "__doc__").splitlines()[0])
            except AttributeError:
                inter_descs.append("")

    headers = ["name", "description"]
    inter_rows = zip(inter_names, inter_descs)

    print_table("New-Style Interfaces", headers, inter_rows)

    print_table("Old-Style Interfaces", ["Name"], zip(list_dev_interfaces()))


def list_interface_plugins(interface_name):
    """List interface plugins."""
    inter = get_interface(interface_name)
    plugins = inter.get_plugins()
    rows = [(pl.name, pl.family, pl.description) for pl in plugins]
    print_table(interface_name, ["Name", "Family", "Description"], rows)


# def print_interface_list(inter):
#     mod_list = inter.get_plugins()
#
#     # Determine column widths for output
#     ncol = len(mod_list[0])
#     headings = ['name', 'type', 'description']
#     if ncol != len(headings):
#         raise ValueError('Number of columns does not match the number of headings')
#     col_widths = [len(head) for head in headings]
#     for mod in mod_list:
#         for col in range(len(mod)):
#             col_widths[col] = max(col_widths[col], len(mod[col]))
#     # col_widths = [wid + 1 for wid in col_widths]
#     print(col_widths)
#
#     # Create and print the header
#     head_vals = []
#     for head_val, width in zip(headings, col_widths):
#         head_vals.append(f'{head_val:{width}}')
#     head_str = ' | '.join(head_vals)
#     print(head_str)
#     print(len(head_str) * '-')
#
#     # Create and print the rows
#     for mod in mod_list:
#         mod_vals = []
#         for mod_val, width in zip(mod, col_widths):
#             mod_vals.append(f'{mod_val:{width}}')
#         mod_str = ' | '.join(mod_vals)
#         print(mod_str)


def main():
    """Command line interface main function."""
    parser = ArgumentParser(description=__doc__, formatter_class=formclass)
    subparsers = parser.add_subparsers(
        title="Commands", description="GeoIPS commands", dest="cmd"
    )

    # run command
    run_parser = subparsers.add_parser(
        "run",
        description="Run a GeoIPS procflow",
        help="Run a GeoIPS procflow",
        formatter_class=formclass,
    )
    run_parser.add_argument("test")

    # list command
    list_parser = subparsers.add_parser(
        "list",
        description="List available interfaces",
        help="List available interfaces",
        formatter_class=formclass,
    )

    # Don't use `dest` here. It doesn't work well with aliases
    # Instead use `set_defaults` on each parser like below
    # See here for more info:
    #    https://bugs.python.org/issue36664
    list_subparsers = list_parser.add_subparsers()

    # list interfaces
    add_list_interface_parser(list_subparsers, "interfaces", aliases=["int", "ints"])

    # list plugins for each interface
    add_list_interface_parser(list_subparsers, "algorithms", aliases=["alg", "algs"])
    # add_list_interface_parser(
    #     list_subparsers, "feature_annotators", aliases=["bound", "bounds"]
    # )
    add_list_interface_parser(
        list_subparsers, "colormappers", aliases=["cmap", "cmaps"]
    )
    add_list_interface_parser(
        list_subparsers, "filename_formatters", aliases=["ff", "ffs"]
    )
    # add_list_interface_parser(
    #     list_subparsers, "gridline_annotators", aliases=["gf", "gfs"]
    # )
    add_list_interface_parser(
        list_subparsers, "interpolators", aliases=["interp", "interps"]
    )
    # add_list_interface_parser(
    #     list_subparsers, "outputter_configs", aliases=["oc", "ocs"]
    # )
    add_list_interface_parser(
        list_subparsers, "output_formatters", aliases=["out", "outs"]
    )
    add_list_interface_parser(list_subparsers, "procflows", aliases=["pf", "pfs"])
    # add_list_interface_parser(list_subparsers, 'products', aliases=['prod', 'prods'])
    add_list_interface_parser(list_subparsers, "readers", aliases=["reader"])
    add_list_interface_parser(
        list_subparsers, "title_formatters", aliases=["tf", "tfs"]
    )

    args = parser.parse_args()

    if args.cmd == "list":
        if args.interface == "interfaces":
            list_interfaces()
        else:
            list_interface_plugins(args.interface)
            # print(get_interface(args.interface))
            # print(get_interface(args.interface).get_plugins())
            # getattr(interfaces, args.interface).get_plugins(
            #     pretty=True, with_family=True, with_description=True
            # )


if __name__ == "__main__":
    main()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read derived surface winds from REMSS SMAP, WINDSAT, and AMSR netcdf data."""

import logging

LOG = logging.getLogger(__name__)

MS_TO_KTS = 1.94384
DEG_TO_KM = 111.321

# interface = None indicates to the GeoIPS interfaces that this is not a valid
# plugin, and this module will not be added to the GeoIPS plugin registry.
# This allows including python modules within the geoips/plugins directory
# that provide helper or utility functions to the geoips plugins, but are
# not full GeoIPS plugins on their own.
interface = None


def read_remss_data(wind_xarray, data_type):
    """
    Reformat SMAP or WindSat xarray object appropriately.

    variables: latitude, longitude, time, wind_speed_kts
    attributes: source_name, platform_name, data_provider,
    interpolation_radius_of_influence
    """
    import xarray
    import numpy
    from datetime import datetime

    # Set attributes appropriately
    if data_type == "smap":
        LOG.info("Reading SMAP data")
        wind_varname = "wind"
        day_of_month_varname = "day_of_month_of observation"
        minute_varname = "minute"
    elif data_type == "windsat":
        LOG.info("Reading WindSat data")
        wind_varname = "wind_speed_TC"
        day_of_month_varname = "day_of_month_of_observation"
        minute_varname = "time"
    elif data_type == "amsr2":
        LOG.info("Reading AMSR2 data")
        wind_varname = "wind_speed_TC"
        day_of_month_varname = "day_of_month_of_observation"
        minute_varname = "time"
    wind_xarray.attrs["full_day_file"] = True

    wind_xarray_1 = xarray.Dataset()
    wind_xarray_1.attrs = wind_xarray.attrs.copy()
    wind_xarray_1.attrs["current_node_dimension"] = "node_dimension = 1"

    wind_xarray_2 = xarray.Dataset()
    wind_xarray_2.attrs = wind_xarray.attrs.copy()
    wind_xarray_2.attrs["current_node_dimension"] = "node_dimension = 2"

    # Set wind_speed_kts appropriately
    winds_1 = numpy.flipud(wind_xarray[wind_varname].values[:, :, 0])
    winds_2 = numpy.flipud(wind_xarray[wind_varname].values[:, :, 1])

    # Full dataset is 720x1440x2, break that up into ascending and descending nodes.
    wind_xarray_1["wind_speed_kts"] = xarray.DataArray(
        winds_1 * MS_TO_KTS, name="wind_speed_kts"
    )
    wind_xarray_1["wind_speed_kts"].attrs = wind_xarray[wind_varname].attrs
    wind_xarray_1["wind_speed_kts"].attrs["units"] = "kts"

    wind_xarray_2["wind_speed_kts"] = xarray.DataArray(
        winds_2 * MS_TO_KTS, name="wind_speed_kts"
    )
    wind_xarray_2["wind_speed_kts"].attrs = wind_xarray[wind_varname].attrs
    wind_xarray_2["wind_speed_kts"].attrs["units"] = "kts"

    # Set lat/lons appropriately
    # These are (1440x720)
    lats2d, lons2d = numpy.meshgrid(wind_xarray.lat, wind_xarray.lon)
    # lats2d = numpy.dstack([lats2d.transpose(), lats2d.transpose()])
    # lons2d = numpy.dstack([lons2d.transpose(), lons2d.transpose()])
    lats2d = lats2d.transpose()
    lons2d = lons2d.transpose()
    wind_xarray_1["latitude"] = xarray.DataArray(data=numpy.flipud(lats2d))
    wind_xarray_1["longitude"] = xarray.DataArray(data=lons2d)
    wind_xarray_2["latitude"] = xarray.DataArray(data=numpy.flipud(lats2d))
    wind_xarray_2["longitude"] = xarray.DataArray(data=lons2d)
    wind_xarray_1 = wind_xarray_1.set_coords(["latitude", "longitude"])
    wind_xarray_2 = wind_xarray_2.set_coords(["latitude", "longitude"])

    from numpy import datetime64

    # Set time appropriately
    year = wind_xarray.year_of_observation
    month = wind_xarray.month_of_observation
    day = wind_xarray.attrs[day_of_month_varname]
    basedt = datetime.strptime(
        "{0:04d}{1:02d}{2:02d}".format(year, month, day), "%Y%m%d"
    )
    # minarr = wind_xarray.minute
    minarr = numpy.flipud(wind_xarray[minute_varname])
    # NOTE there is a version of numpy 2.x that will break for masked datetime64
    # arrays.  The latest stable version of numpy 2.2.4 works without modification,
    # but beware that some versions were broken.
    timearr = datetime64(basedt) + minarr
    wind_xarray_1["time"] = xarray.DataArray(timearr[:, :, 0])
    wind_xarray_2["time"] = xarray.DataArray(timearr[:, :, 1])
    wind_xarray_1 = wind_xarray_1.set_coords(["time"])
    wind_xarray_2 = wind_xarray_2.set_coords(["time"])
    return {"WINDSPEED_1": wind_xarray_1, "WINDSPEED_2": wind_xarray_2}



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Command line script for updating the TC database."""


def main():
    """Update TC tracks database via command line."""
    import sys
    from geoips.commandline.log_setup import setup_logging

    LOG = setup_logging()
    from geoips.sector_utils.tc_tracks_database import check_db

    if len(sys.argv) > 2:
        LOG.info("Calling with arg %s", sys.argv[1:])
        check_db(sys.argv[1:])
    else:
        check_db()


if __name__ == "__main__":
    main()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Passive Microwave 89 GHz Colorized Brightness Temperature.

Data manipulation steps for the "color89" product.
This algorithm expects Brightness Temperatures in units of degrees Kelvin
"""
import logging

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "list_numpy_to_numpy"
name = "pmw_color89"


def call(arrays):
    """color89 product algorithm data manipulation steps.

    This algorithm expects Brightness Temperatures in units of degrees Kelvin,
    and returns red green and blue gun arrays.

    Parameters
    ----------
    arrays : list of numpy.ndarray
        * list of numpy.ndarray or numpy.MaskedArray of channel data and
            other variables, in order of sensor "variables" list
        * Channel data: Degrees Kelvin

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of qualitative RGBA image output
    """
    h89 = arrays[0]
    v89 = arrays[1]

    from geoips.data_manipulations.corrections import apply_data_range, apply_gamma

    red = 1.818 * v89 - 0.818 * h89
    red = apply_data_range(
        red,
        220.0,
        310.0,
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=True,
    )
    red = apply_gamma(red, 1.0)

    grn = (h89 - 240.0) / (300.0 - 240.0)
    grn = apply_data_range(
        grn,
        0.0,
        1.0,
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )
    grn = apply_gamma(grn, 1.0)

    blu = (v89 - 270.0) / (290.0 - 270.0)
    blu = apply_data_range(
        blu,
        0.0,
        1.0,
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )
    blu = apply_gamma(blu, 1.0)

    from geoips.image_utils.mpl_utils import alpha_from_masked_arrays, rgba_from_arrays

    alp = alpha_from_masked_arrays([red, grn, blu])
    rgba = rgba_from_arrays(red, grn, blu, alp)
    return rgba



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `list plugins` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

from importlib import resources
import json
import pytest

from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest


class TestGeoipsListPlugins(BaseCliTest):
    """Unit Testing Class for List Plugins Command."""

    @property
    def command_combinations(self):
        """A list of every possible call signature for the GeoipsListPlugins command.

        This includes failing cases as well.
        """
        if not hasattr(self, "_cmd_list"):
            base_args = ["geoips", "list", "plugins"]
            alias_args = ["geoips", "ls", "plugins"]
            self._cmd_list = [base_args, alias_args]
            for argset in [base_args, alias_args]:
                for pkg_name in self.plugin_package_names:
                    self._cmd_list.append(argset + ["-p", pkg_name])
            # Add argument list invoking the --columns flag
            self._cmd_list.append(base_args + ["--columns", "package", "interface"])
            self._cmd_list.append(base_args + ["--columns", "plugin_type", "family"])
            self._cmd_list.append(base_args + ["--columns", "relpath", "source_names"])
            self._cmd_list.append(base_args + ["--columns", "plugin_name", "relpath"])
            self._cmd_list.append(alias_args + ["--columns", "package", "interface"])
            self._cmd_list.append(alias_args + ["--columns", "plugin_type", "family"])
            self._cmd_list.append(alias_args + ["--columns", "relpath", "source_names"])
            self._cmd_list.append(alias_args + ["--columns", "plugin_name", "relpath"])
            # Add argument list which invokes the help message for this command
            self._cmd_list.append(base_args + ["-h"])
            self._cmd_list.append(alias_args + ["-h"])
            # Add argument list with a non-existent package name
            self._cmd_list.append(base_args + ["-p", "non_existent_package"])
            self._cmd_list.append(alias_args + ["-p", "non_existent_package"])
        return self._cmd_list

    def check_error(self, args, error):
        """Ensure that the 'geoips list plugins ...' error output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        # bad command has been provided, check the contents of the error message
        assert args != ["geoips", "list", "plugins"] and args != [
            "geoips",
            "ls",
            "plugins",
        ]
        assert "usage: To use, type `geoips list plugins`" in error

    def check_output(self, args, output):
        """Ensure that the 'geoips list plugins ...' successful output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        if "usage: To use, type" in output:
            # -h has been called, check help message contents for this command
            assert "-h" in args
            assert "To use, type `geoips list plugins`" in output
        else:
            # The args provided are valid, so test that the output is actually correct
            selected_cols = self.retrieve_selected_columns(args)
            headers = {
                "GeoIPS Package": "package",
                "Interface Name": "interface",
                "Interface Type": "plugin_type",
                "Family": "family",
                "Plugin Name": "plugin_name",
                "Relative Path": "relpath",
            }
            # Assert that the correct headers exist in the CLI output
            self.assert_correct_headers_in_output(output, headers, selected_cols)

            if "-p" in args:
                # certain package has been selected, ensure we found every pkg-plugin
                pkg_names = [args[-1]]
            else:
                # all packages selected, ensure that we found every plugin
                pkg_names = self.plugin_package_names
            for pkg_name in pkg_names:
                with open(
                    str(resources.files(pkg_name) / "registered_plugins.json"), "r"
                ) as fo:
                    plugin_registry = json.load(fo)
                for interface_type in ["module_based", "yaml_based"]:
                    # check each type of interface
                    if interface_type in plugin_registry:
                        # if that interface type is within the registry, loop over the
                        # found interfaces
                        for interface_name in plugin_registry[interface_type]:
                            # search the interface's portion of the registry and assert
                            # that each plugin found within it has been reported in the
                            # CLI"s output
                            interface_registry = plugin_registry[interface_type][
                                interface_name
                            ]
                            for plugin_name in interface_registry:
                                assert plugin_name in list(interface_registry.keys())


test_sub_cmd = TestGeoipsListPlugins()


@pytest.mark.parametrize(
    "args",
    test_sub_cmd.command_combinations,
    ids=test_sub_cmd.generate_id,
)
def test_command_combinations(monkeypatch, args):
    """Test all 'geoips list plugins ...' commands.

    This test covers every valid combination of commands for the
    'geoips list plugins' command. We also test invalid commands, to ensure that
    the proper help documentation  is provided for those using the command incorrectly.

    Parameters
    ----------
    args: 2D array of str
        - List of arguments to call the CLI with (ie. ['geoips', 'list', 'plugins'])
    """
    test_sub_cmd.test_command_combinations(monkeypatch, args)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Interpolators interface module."""

from geoips.interfaces.base import BaseModuleInterface


class CoverageCheckersInterface(BaseModuleInterface):
    """Interpolation routine to apply when reprojecting data."""

    name = "coverage_checkers"
    required_args = {"standard": ["xarray_obj", "variable_name"]}
    required_kwargs = {"standard": {}}
    allowable_kwargs = {
        "standard": {
            "area_def",
            "radius_km",
        }
    }

    def get_plugin_for_product(self, product, checker_field="coverage_checker"):
        """Get plugin for product."""
        if checker_field in product:
            self.get_plugin(product[checker_field]["name"])
        else:
            self.get_plugin("masked_array")


coverage_checkers = CoverageCheckersInterface()
